# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/write_mass_file.py
#
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports  
from RCAIDE.Library.Methods.Aerodynamics.Athena_Vortex_Lattice.purge_files import purge_files
from RCAIDE.Library.Methods.Aerodynamics.Athena_Vortex_Lattice.create_avl_datastructures import translate_avl_wing, translate_avl_body

# ----------------------------------------------------------------------------------------------------------------------
#  write_geometry
# ---------------------------------------------------------------------------------------------------------------------- 
def write_geometry(avl_object,run_script_path):
    """This function writes the translated aircraft geometry into text file read 
    by AVL when it is called

    Assumptions:
        None
        
    Source:
        Drela, M. and Youngren, H., AVL, http://web.mit.edu/drela/Public/web/avl

    Inputs:
        avl_object

    Outputs:
        None

    Properties Used:
        N/A
    """    
    
    # unpack inputs
    aircraft                          = avl_object.vehicle
    geometry_file                     = avl_object.settings.filenames.features
    number_of_spanwise_vortices       = avl_object.settings.number_of_spanwise_vortices
    number_of_chordwise_vortices      = avl_object.settings.number_of_chordwise_vortices 
    
    avl_object.reference_values.S_ref        = avl_object.vehicle.wings['main_wing'].areas.reference
    avl_object.reference_values.c_ref        = avl_object.vehicle.wings['main_wing'].chords.mean_aerodynamic
    avl_object.reference_values.b_ref        = avl_object.vehicle.wings['main_wing'].spans.projected
    avl_object.reference_values.X_ref        = avl_object.vehicle.mass_properties.center_of_gravity[0][0]
    avl_object.reference_values.Y_ref        = avl_object.vehicle.mass_properties.center_of_gravity[0][1]
    avl_object.reference_values.Z_ref        = avl_object.vehicle.mass_properties.center_of_gravity[0][2]
    avl_object.reference_values.aspect_ratio = (avl_object.reference_values.b_ref ** 2) / avl_object.reference_values.S_ref 
    
    # Open the geometry file after purging if it already exists
    purge_files([geometry_file]) 
    geometry             = open(geometry_file,'w')

    with open(geometry_file,'w') as geometry:
        header_text       = make_header_text(avl_object)
        geometry.write(header_text)
        
        for w in aircraft.wings:
            avl_wing      = translate_avl_wing(w)
            wing_text     = make_surface_text(avl_wing,number_of_spanwise_vortices,number_of_chordwise_vortices)
            geometry.write(wing_text)  
        
        if avl_object.settings.model_fuselage:
            for b in aircraft.fuselages:
                avl_body  = translate_avl_body(b)
                body_text = make_body_text(avl_body,number_of_chordwise_vortices)
                geometry.write(body_text)
            
    return


def make_header_text(avl_object):  
    """This function writes the header using the template required for the AVL executable to read

    Assumptions:
        None
        
    Source:
        None

    Inputs:
        avl_object.settings.flow_symmetry.xz_plane                      [-]
        avl_object.settings.flow_symmetry.xy_parallel                   [-]
        avl_object.settings.flow_symmetry.z_symmetry_plane              [-]
        avl_object.vehicle.wings['main_wing'].areas.reference          [meters**2]
        avl_object.vehicle.wings['main_wing'].chords.mean_aerodynamic  [meters]
        avl_object.vehicle.wings['main_wing'].spans.projected          [meters]
        avl_object.vehicle.mass_properties.center_of_gravity           [meters]
        avl_object.vehicle.tag                                         [-]
    
    Outputs:
        header_text                                                     [-]

    Properties Used:
        N/A
    """      
    header_base = \
'''{0}

#Mach
 {1}
 
#Iysym   IZsym   Zsym
  {2}      {3}     {4}
  
#Sref    Cref    Bref 	<meters>
{5}      {6}     {7}

#Xref    Yref    Zref   <meters>
{8}      {9}     {10}

'''

    # Unpack inputs
    Iysym = avl_object.settings.flow_symmetry.xz_plane
    Izsym = avl_object.settings.flow_symmetry.xy_parallel
    Zsym  = avl_object.settings.flow_symmetry.z_symmetry_plane
    Sref  = avl_object.vehicle.wings['main_wing'].areas.reference
    Cref  = avl_object.vehicle.wings['main_wing'].chords.mean_aerodynamic
    Bref  = avl_object.vehicle.wings['main_wing'].spans.projected
    Xref  = avl_object.vehicle.mass_properties.center_of_gravity[0][0]
    Yref  = avl_object.vehicle.mass_properties.center_of_gravity[0][1]
    Zref  = avl_object.vehicle.mass_properties.center_of_gravity[0][2]
    name  = avl_object.vehicle.tag     
    mach = 0.0

    # Insert inputs into the template
    header_text = header_base.format(name,mach,Iysym,Izsym,Zsym,Sref,Cref,Bref,Xref,Yref,Zref)

    return header_text


def make_surface_text(avl_wing,number_of_spanwise_vortices,number_of_chordwise_vortices):
    """This function writes the surface text using the template required for the AVL executable to read

    Assumptions:
        None
        
    Source:
        None

    Inputs:
       avl_wing.symmetric
       avl_wing.tag
        
    Outputs:
        surface_text                                                 

    Properties Used:
        N/A
    """       
    ordered_tags = []         
    surface_base = \
        '''

#---------------------------------------------------------
SURFACE
{0}
#Nchordwise  Cspace   Nspanwise  Sspace
{1}         {2}         {3}      {4}{5}
'''        
    # Unpack inputs
    symm = avl_wing.symmetric
    name = avl_wing.tag

    if symm:
        ydup = '\n\nYDUPLICATE\n0.0\n' # Duplication of wing about xz plane
    else:
        ydup     = ' ' 
    
    # Vertical Wings
    if avl_wing.vertical:
        # Define precision of analysis. See AVL documentation for reference 
        chordwise_vortex_spacing = 1.0
        spanwise_vortex_spacing  = -1.1                              # cosine distribution i.e. || |   |    |    |  | ||
        ordered_tags = sorted(avl_wing.sections, key = lambda x: x.origin[0][2])
        
        # Write text 
        surface_text = surface_base.format(name,number_of_chordwise_vortices,chordwise_vortex_spacing,number_of_spanwise_vortices ,spanwise_vortex_spacing,ydup)
        for i in range(len(ordered_tags)):
            section_text    = make_wing_section_text(ordered_tags[i])
            surface_text    = surface_text + section_text
            
    # Horizontal Wings        
    else:        
        # Define precision of analysis. See AVL documentation for reference
        chordwise_vortex_spacing = 1.0        
        spanwise_vortex_spacing  = 1.0                              # cosine distribution i.e. || |   |    |    |  | ||
        ordered_tags = sorted(avl_wing.sections, key = lambda x: x.origin[0][1])
    
        # Write text  
        surface_text = surface_base.format(name,number_of_chordwise_vortices,chordwise_vortex_spacing,number_of_spanwise_vortices ,spanwise_vortex_spacing,ydup)
        for i in range(len(ordered_tags)):
            section_text    = make_wing_section_text(ordered_tags[i])
            surface_text    = surface_text + section_text

    return surface_text


def make_body_text(avl_body,number_of_chordwise_vortices):   
    """This function writes the body text using the template required for the AVL executable to read

    Assumptions:
        None
        
    Source:
        None

    Inputs:
        avl_body.sections.horizontal
        avl_body.sections.vertical
    
    Outputs:
        body_text                                                 

    Properties Used:
        N/A
    """      
    surface_base = \
'''

#---------------------------------------------------------
SURFACE
{0}
#Nchordwise  Cspace   Nspanwise  Sspace
{1}           {2}      
'''
    # Unpack inputs
    name = avl_body.tag
    
    # Define precision of analysis. See AVL documentation for reference 
    chordwise_vortex_spacing = 1.0 
    
    # Form the horizontal part of the + shaped fuselage    
    hname           = name + '_horizontal'
    horizontal_text = surface_base.format(hname,number_of_chordwise_vortices,chordwise_vortex_spacing)
       
    ordered_tags = []
    ordered_tags = sorted(avl_body.sections.horizontal, key = lambda x: x.origin[1])
    for i in range(len(ordered_tags)):
        section_text    = make_body_section_text(ordered_tags[i])
        horizontal_text = horizontal_text + section_text
        
    # Form the vertical part of the + shaped fuselage
    vname         = name + '_vertical'
    vertical_text = surface_base.format(vname,number_of_chordwise_vortices,chordwise_vortex_spacing)   
    ordered_tags = []
    ordered_tags = sorted(avl_body.sections.vertical, key = lambda x: x.origin[2])
    for i in range(len(ordered_tags)):
        section_text    = make_body_section_text(ordered_tags[i])
        vertical_text = vertical_text + section_text
        
    body_text = horizontal_text + vertical_text
    return body_text  


def make_wing_section_text(avl_section):
    """This function writes the wing text using the template required for the AVL executable to read

    Assumptions:
        None
        
    Source:
        None

    Inputs:
       avl_section.origin             [meters]
       avl_section.chord              [meters]
       avl_section.twist              [radians]
       avl_section.airfoil_coord_file [-] 
        
    Outputs:
        wing_section_text                                                 

    Properties Used:
        N/A
    """      
    section_base = \
'''
SECTION
#Xle    Yle      Zle      Chord     Ainc  Nspanwise  Sspace
{0}  {1}    {2}    {3}    {4}     
'''
    airfoil_base = \
'''AFILE
{}
'''
    naca_airfoil_base = \
'''NACA
{}
'''
    # Unpack inputs
    x_le          = avl_section.origin[0][0]
    y_le          = avl_section.origin[0][1]
    z_le          = avl_section.origin[0][2]
    chord         = avl_section.chord
    ainc          = avl_section.twist
    airfoil_coord = avl_section.airfoil_coord_file
    naca_airfoil  = avl_section.naca_airfoil 
     
    wing_section_text = section_base.format(round(x_le,4),round(y_le,4), round(z_le,4),round(chord,4),round(ainc,4))
    if airfoil_coord:
        wing_section_text = wing_section_text + airfoil_base.format(airfoil_coord)
    if naca_airfoil:
        wing_section_text = wing_section_text + naca_airfoil_base.format(naca_airfoil)        
    
    ordered_cs = []
    ordered_cs = sorted(avl_section.control_surfaces, key = lambda x: x.order)
    for i in range(len(ordered_cs)):
        control_text = make_controls_text(ordered_cs[i])
        wing_section_text = wing_section_text + control_text

    return wing_section_text

    
def make_body_section_text(avl_body_section):
    """This function writes the body text using the template required for the AVL executable to read

    Assumptions:
        None
        
    Source:
        None

    Inputs:
       avl_section.origin             [meters]
       avl_section.chord              [meters]
       avl_section.twist              [radians]
       avl_section.airfoil_coord_file [-] 
                  
    Outputs:
        body_section_text                                                 

    Properties Used:
        N/A
    """    
    section_base = \
'''
SECTION
#Xle     Yle      Zle      Chord     Ainc  Nspanwise  Sspace
{0}    {1}     {2}     {3}     {4}      1        0
'''
    airfoil_base = \
'''AFILE
{}
'''

    # Unpack inputs
    x_le    = avl_body_section.origin[0]
    y_le    = avl_body_section.origin[1]
    z_le    = avl_body_section.origin[2]
    chord   = avl_body_section.chord
    ainc    = avl_body_section.twist
    airfoil = avl_body_section.airfoil_coord_file

    body_section_text = section_base.format(round(x_le,4),round(y_le,4), round(z_le,4),round(chord,4),round(ainc,4))
    if airfoil:
        body_section_text = body_section_text + airfoil_base.format(airfoil)
    
    return body_section_text

    
def make_controls_text(avl_control_surface):
    """This function writes the control surface text using the template required 
    for the AVL executable to read

    Assumptions:
        None
        
    Source:
        None

    Inputs:
        avl_control_surface.tag             [-]
        avl_control_surface.gain            [-]
        avl_control_surface.x_hinge         [-]
        avl_control_surface.hinge_vector    [-]
        avl_control_surface.sign_duplicate  [-]
                  
    Outputs:
        control_text                                                 

    Properties Used:
        N/A
    """    
    control_base = \
'''CONTROL
{0}    {1}   {2}   {3}  {4}
'''

    # Unpack inputs
    name     = avl_control_surface.tag
    gain     = avl_control_surface.gain
    xhinge   = avl_control_surface.x_hinge
    hv       = avl_control_surface.hinge_vector
    sign_dup = avl_control_surface.sign_duplicate

    control_text = control_base.format(name,gain,xhinge,hv,sign_dup)

    return control_text



# RCAIDE/Methods/Energy/Propulsors/Turboprop_Propulsor/size_core.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Library.Methods.Propulsors.Turboprop_Propulsor.compute_thrust import compute_thrust 

# ----------------------------------------------------------------------------------------------------------------------
#  size_core
# ---------------------------------------------------------------------------------------------------------------------- 
def size_core(turboprop,turboprop_conditions,conditions):
    """Sizes the core flow for the design condition.

    Assumptions:
    Perfect gas
    Turboprop Engine

    Sources:
    [1] 

    Inputs:
    conditions.freestream.speed_of_sound [m/s] (conditions is also passed to turboprop.compute(..))
    turboprop.inputs.
      bypass_ratio                            [-]
      total_temperature_reference             [K]
      total_pressure_reference                [Pa]
      number_of_engines                       [-]

    Outputs:
    turboprop.outputs.non_dimensional_power  [-]

    Properties Used:
    turboprop.
      reference_temperature                   [K]
      reference_pressure                      [Pa]
      total_design                            [W] - Design power
    """             

    compute_thrust(turboprop,turboprop_conditions,conditions)  
       
    return    



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_payload_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units ,  Data
  
# ----------------------------------------------------------------------------------------------------------------------
#  Operating Items Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_payload_weight(vehicle, weight_per_passenger = 165. * Units.lb):
    """ Calculate the payload weight, including:
        - passenger and carry-on weight
        - baggage weight
        - cargo weight

        Assumptions:
            None

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.passengers: number of passengers in aircraft
                -.design_range: design range of aircraft                        [nmi]
                -.mass_properties.cargo: weight of cargo carried                [kilograms]

        Outputs:
            output - data dictionary with weights                               [kilograms]
                    - output.passengers: passenger weight
                    - output.baggage: baggage weight
                    - output.cargo: cargo weight
                    - output.total: total payload weight

        Properties Used:
            N/A
    """
    WPPASS  = weight_per_passenger
    WPASS   = vehicle.passengers * WPPASS
    DESRNG  = vehicle.flight_envelope.design_range / Units.nmi
    if DESRNG <= 900:
        BPP = 35 * Units.lbs  # luggage weight per passenger depends on the design range
    elif DESRNG <= 2900:
        BPP = 40 * Units.lbs
    else:
        BPP = 44 * Units.lbs
    WPBAG       = BPP * vehicle.passengers  # baggage weight
    WPAYLOAD    = WPASS + WPBAG + vehicle.mass_properties.cargo / Units.lbs  # payload weight

    output              = Data()
    output.total        = WPAYLOAD
    output.passengers   = WPASS
    output.baggage      = WPBAG
    output.cargo        = vehicle.mass_properties.cargo
    return output



# RCAIDE/Library/Methods/Propulsors/Constant_Speed_ICE_Propulsor/append_ice_cs_propeller_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_ice_cs_propeller_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_ice_cs_propeller_conditions(ice_cs_propeller,segment):  
    ones_row    = segment.state.ones_row                  
    segment.state.conditions.energy[ice_cs_propeller.tag]                               = Conditions()  
    segment.state.conditions.energy[ice_cs_propeller.tag].throttle                      = 0. * ones_row(1)      
    segment.state.conditions.energy[ice_cs_propeller.tag].commanded_thrust_vector_angle = 0. * ones_row(1)  
    segment.state.conditions.energy[ice_cs_propeller.tag].thrust                        = 0. * ones_row(3) 
    segment.state.conditions.energy[ice_cs_propeller.tag].power                         = 0. * ones_row(1) 
    segment.state.conditions.energy[ice_cs_propeller.tag].moment                        = 0. * ones_row(3) 
    segment.state.conditions.energy[ice_cs_propeller.tag].fuel_flow_rate                = 0. * ones_row(1)
    segment.state.conditions.energy[ice_cs_propeller.tag].rpm                           = segment.state.conditions.energy.rpm * ones_row(1)      
    segment.state.conditions.energy[ice_cs_propeller.tag].inputs                        = Conditions()
    segment.state.conditions.energy[ice_cs_propeller.tag].outputs                       = Conditions() 
    segment.state.conditions.noise[ice_cs_propeller.tag]                                = Conditions()  
    return 


# RCAIDE/Library/Components/Propulsors/Converters/Fan.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Feb 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports   
from RCAIDE.Library.Components                      import Component  
from RCAIDE.Library.Methods.Propulsors.Converters.Fan.append_fan_conditions import append_fan_conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  Fan  
# ----------------------------------------------------------------------------------------------------------------------
class Fan(Component):
    """
    A fan component model for turbofan and ducted fan propulsion systems.

    Attributes
    ----------
    tag : str
        Identifier for the fan component. Default is 'Fan'.
        
    polytropic_efficiency : float
        Efficiency of the compression process accounting for losses. Default is 1.0.
        
    pressure_ratio : float
        Ratio of outlet to inlet total pressure. Default is 1.0.
        
    angular_velocity : float
        Rotational speed of the fan [rad/s]. Default is 0.0.

    Notes
    -----
    The Fan class models the compression and energy addition process in a fan stage.
    The model:
    * Calculates work input required for given pressure ratio
    * Accounts for losses through polytropic efficiency
    * Handles variable speed operation
    * Assumes axial flow conditions
    * Models both subsonic and transonic fan operation

    **Major Assumptions**
    * Pressure ratio and efficiency do not change with varying conditions
    * Uniform flow at inlet and exit
    * No radial variations in flow properties
    * Adiabatic process (no heat transfer with surroundings)
    * Steady flow conditions

    **Definitions**

    'Polytropic Efficiency'
        Measure of compression efficiency accounting for real gas effects

    'Pressure Ratio'
        Ratio of exit to inlet total pressure
        
    'Angular Velocity'
        Rotational speed of the fan rotor

    References
    ----------
    [1] Mattingly, J. D., & Boyer, K. M. (2016). Elements of propulsion: Gas 
        turbines and rockets, second edition Jack D. Mattingly, Keith M. Boyer. 
        American Institute of Aeronautics and Astronautics.

    See Also
    --------
    RCAIDE.Library.Components.Component
    RCAIDE.Library.Methods.Propulsors.Converters.Fan.append_fan_conditions
    """
    
    def __defaults__(self):
        """This sets the default values for the component to function.

        Assumptions:
            Pressure ratio and efficiency do not change with varying conditions.

        Source:
            https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/ 
        """         
        #set the default values
        self.tag                            = 'Fan'
        self.polytropic_efficiency          = 1.0
        self.pressure_ratio                 = 1.0
        self.angular_velocity               = 0


    def append_operating_conditions(self,segment,propulsor): 
        propulsor_conditions =  segment.state.conditions.energy[propulsor.tag]
        append_fan_conditions(self,segment,propulsor_conditions)
        return                                


# RCAIDE/Methods/Energy/Propulsors/Turboprop_Propulsor/compute_thrust.py
# 
# 
# Created:  Sep 2024, M. Clarke, M. Guidotti

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports  
from RCAIDE.Framework.Core      import Units 

# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  compute_thrust
# ----------------------------------------------------------------------------------------------------------------------
def compute_thrust(turboprop,turboprop_conditions,conditions):
    
    """Computes thrust and other properties as below.

    Assumptions:
    Perfect gas

    Source:
    Mattingly, Jack D.. “Elements of Gas Turbine Propulsion.” (1996).

    Inputs:
    conditions.freestream.gravity 
    conditions.freestream.temperature
    conditions.freestream.pressure      
    conditions.freestream.mach_number
    conditions.freestream.velocity
    conditions.freestream.speed_of_sound
    turboprop.reference_temperature
    turboprop.reference_pressure 
    turboprop.combustor.turbine_inlet_temperature
    turboprop.design_propeller_efficiency
    turboprop.design_gearbox_efficiency
    turboprop.low_pressure_turbine.mechanical_efficiency
    turboprop.combustor.fuel_data.lower_heating_value
    turboprop_conditions.stag_temp_hpt_out
    turboprop_conditions.stag_temp_hpt_in
    turboprop_conditions.stag_temp_lpt_out
    turboprop_conditions.stag_temp_lpt_in
    turboprop_conditions.total_temperature_reference
    turboprop_conditions.total_pressure_reference     
    turboprop_conditions.fuel_to_air_ratio 
    turboprop_conditions.cpt
    turboprop_conditions.cpc
    turboprop_conditions.R_t
    turboprop_conditions.R_c
    turboprop_conditions.T9
    turboprop_conditions.P9  
    turboprop_conditions.gamma_c
    turboprop_conditions.core_exit_velocity
    
    Outputs:
    turboprop_conditions.thrust                           
    turboprop_conditions.thrust_specific_fuel_consumption 
    turboprop_conditions.non_dimensional_thrust           
    turboprop_conditions.core_mass_flow_rate              
    turboprop_conditions.fuel_flow_rate                   
    turboprop_conditions.power                            
    turboprop_conditions.specific_power                   
    turboprop_conditions.power_specific_fuel_consumption  
    turboprop_conditions.thermal_efficiency               
    turboprop_conditions.propulsive_efficiency               
    """       
    
    #compute dimensional mass flow rates 
    g                                              = conditions.freestream.gravity                                                                             # [m/s**2]
    Tref                                           = turboprop.reference_temperature                                                                           # [K]
    Pref                                           = turboprop.reference_pressure                                                                              # [Pa]
    total_temperature_reference                    = turboprop_conditions.total_temperature_reference                                                          # [K]
    total_pressure_reference                       = turboprop_conditions.total_pressure_reference                                                             # [Pa]
                                                                                                                                                                          
    #unpack from turboprop                                                                                                                                     
    f                                              = turboprop_conditions.fuel_to_air_ratio                                                                    # [-]
    cp_t                                           = turboprop_conditions.cpt                                                                                  # [J/kg*K]
    cp_c                                           = turboprop_conditions.cpc                                                                                  # [J/kg*K]   
    R_t                                            = turboprop_conditions.R_t                                                                                  # [J/mol*K]    
    R_c                                            = turboprop_conditions.R_c                                                                                  # [J/mol*K]
    T9                                             = turboprop_conditions.T9                                                                                   # [K]
    P9                                             = turboprop_conditions.P9                                                                                   # [Pa]     
    gamma_c                                        = turboprop_conditions.gamma_c                                                                              # [-]
    V9                                             = turboprop_conditions.core_exit_velocity                                                                   # [m/s]
    T0                                             = conditions.freestream.temperature                                                                         # [K]
    P0                                             = conditions.freestream.pressure                                                                            # [Pa]   
    M0                                             = conditions.freestream.mach_number                                                                         # [-]    
    V0                                             = conditions.freestream.velocity                                                                            # [m/s]
    a0                                             = conditions.freestream.speed_of_sound                                                                      # [m/s]
    Tt4                                            = turboprop.combustor.turbine_inlet_temperature                                                             # [K]     
    eta_prop                                       = turboprop.design_propeller_efficiency                                                                     # [-]
    eta_g                                          = turboprop.design_gearbox_efficiency                                                                       # [-]
    eta_mL                                         = turboprop.low_pressure_turbine.mechanical_efficiency                                                      # [-]
    h_PR                                           = turboprop.combustor.fuel_data.lower_heating_value                                                         # [J/kg]
    tau_tH                                         = (turboprop_conditions.stag_temp_hpt_out/turboprop_conditions.stag_temp_hpt_in)                            # [-]
    tau_tL                                         = (turboprop_conditions.stag_temp_lpt_out/turboprop_conditions.stag_temp_lpt_in)                            # [-]

    C_prop                                         = eta_prop*eta_g*eta_mL*(1 + f)*(cp_t*Tt4)/(cp_c*T0)*tau_tH*(1 - tau_tL)                                    # [-]
    Cc                                             = (gamma_c - 1)*M0*((1 + f)*(V9/a0) - M0 + (1 + f)*(R_t/R_c)*((T9/T0)/((V9/a0)))*((1 - (P0/P9))/gamma_c))   # [-]
    C_tot                                          = Cc + C_prop                                                                                               # [-]
    Fsp                                            = (C_tot*cp_c*T0)/(V0)                                                                                      # [(N*s)/kg] 
    TSFC                                           = (f/(Fsp)) * Units.hour                                                                                    # [kg/(N*hr)] 
    W_dot_mdot0                                    = C_tot*cp_c*T0                                                                                             # [(W*s)/kg] 
    PSFC                                           = (f/(C_tot*cp_c*T0)) * Units.hour                                                                          # [kg/(W*hr)]
    eta_T                                          = C_tot/((f*h_PR)/(cp_c*T0))                                                                                # [-]
    eta_P                                          = C_tot/((C_prop/eta_prop) + ((gamma_c - 1)/2)*((1 + f)*((V9/a0))**2 - M0**2))                              # [-]   
    
    mdot_core                                      = turboprop.design_thrust*turboprop_conditions.throttle/(Fsp)                                               # [kg/s]
    mdhc                                           = mdot_core/ (np.sqrt(Tref/total_temperature_reference)*(total_pressure_reference/Pref))                    # [kg/s]
    
    #computing the dimensional thrust
    FD2                                            = Fsp*mdot_core                                                                                             # [N]  

    #fuel flow rate
    a                                              = np.array([0.])                                                                                            # [-]     
    fuel_flow_rate                                 = np.fmax(FD2*TSFC/g,a)*1./Units.hour                                                                       # [kg/s]  

    #computing the power 
    power                                          = FD2*V0

    # pack outputs 
    turboprop_conditions.thrust                            = FD2 
    turboprop_conditions.thrust_specific_fuel_consumption  = TSFC
    turboprop_conditions.non_dimensional_thrust            = Fsp 
    turboprop_conditions.core_mass_flow_rate               = mdot_core
    turboprop_conditions.fuel_flow_rate                    = fuel_flow_rate    
    turboprop_conditions.power                             = power  
    turboprop_conditions.specific_power                    = W_dot_mdot0  
    turboprop_conditions.power_specific_fuel_consumption   = PSFC 
    turboprop_conditions.thermal_efficiency                = eta_T
    turboprop_conditions.propulsive_efficiency             = eta_P

    return 


# RCAIDE/Framework/Analyses/Noise/Correlation_Buildup.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE Imports 
from RCAIDE.Library.Methods.Noise.Correlation_Buildup.Airframe.airframe_noise         import airframe_noise
from RCAIDE.Library.Methods.Noise.Correlation_Buildup.Turbofan.turbofan_engine_noise  import turbofan_engine_noise   
from RCAIDE.Library.Methods.Noise.Common.decibel_arithmetic                           import SPL_arithmetic  
from RCAIDE.Library.Methods.Noise.Common.generate_hemisphere_microphone_locations     import generate_hemisphere_microphone_locations  
from .Noise      import Noise   

# package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Correlation_Buildup
# ---------------------------------------------------------------------------------------------------------------------- 
class Correlation_Buildup(Noise): 
    """This is an acoustic analysis based on a collection of correlative modes

     Assumptions: 
 
     Source:
     N/A
 
     Inputs:
     None
 
     Outputs:
     None
 
     Properties Used:
     N/A 
    """    
    
    def __defaults__(self):
        
        """ This sets the default values for the analysis.
        
            Assumptions:
            Ground microphone angles start in front of the aircraft (0 deg) and sweep in a lateral direction 
            to the starboard wing and around to the tail (180 deg)
            
            Source:
            N/A
            
            Inputs:
            None
            
            Output:
            None
            
            Properties Used:
            N/A
        """
        
        # Initialize quantities
        self.tag =  "Correlation_Buildup"
        return
            
    def evaluate_noise(self,segment):
        """ Process vehicle to setup geometry, condititon and configuration
    
        Assumptions:
        None
    
        Source:
        N/4
    
        Inputs:
        self.settings.
            center_frequencies  - 1/3 octave band frequencies   [unitless]
    
        Outputs:
        None
    
        Properties Used: 
        """         
    
        # unpack 
        config        = segment.analyses.noise.vehicle 
        settings      = self.settings     
        conditions    = segment.state.conditions  
        dim_cf        = len(settings.center_frequencies ) 
        ctrl_pts      = int(segment.state.numerics.number_of_control_points) 
         
        microphone_locations = generate_hemisphere_microphone_locations(settings)      
        N_hemisphere_mics    = len(microphone_locations)
        
        # create empty arrays for results      
        total_SPL_dBA        = np.ones((ctrl_pts,N_hemisphere_mics))*1E-16 
        total_SPL_spectra    = np.ones((ctrl_pts,N_hemisphere_mics,dim_cf))*1E-16
    
        # flap noise - only applicable for turbofan aircraft
        if 'flap' in config.wings.main_wing.control_surfaces:    
            airframe_noise_res        = airframe_noise(microphone_locations,segment,config,settings) 
            total_SPL_dBA             = SPL_arithmetic(np.concatenate((total_SPL_dBA[:,None,:],airframe_noise_res.SPL_dBA[:,None,:]),axis =1),sum_axis=1)
            total_SPL_spectra[:,:,5:] = SPL_arithmetic(np.concatenate((total_SPL_spectra[:,None,:,5:],airframe_noise_res.SPL_1_3_spectrum[:,None,:,:]),axis =1),sum_axis=1)
              
          # iterate through sources  
        for network in config.networks:  
            for propulsor in network.propulsors:        
                engine_noise                = turbofan_engine_noise(microphone_locations,propulsor,conditions.noise[propulsor.tag].turbofan,segment,settings)    
                total_SPL_dBA               = SPL_arithmetic(np.concatenate((total_SPL_dBA[:,None,:],engine_noise.SPL_dBA[:,None,:]),axis =1),sum_axis=1)
                total_SPL_spectra[:,:,5:]   = SPL_arithmetic(np.concatenate((total_SPL_spectra[:,None,:,5:],engine_noise.SPL_1_3_spectrum[:,None,:,:]),axis =1),sum_axis=1)
                     
        conditions.noise.hemisphere_SPL_dBA              = total_SPL_dBA
        conditions.noise.hemisphere_SPL_1_3_spectrum_dBA = total_SPL_spectra                                                      
        return   




# RCAIDE/Methods/Energy/Sources/Battery/Common/find_mass_gain_rate.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  METHOD
# ---------------------------------------------------------------------------------------------------------------------- -Common
def find_mass_gain_rate(battery,power):
    """finds the mass gain rate of the battery from the ambient air
    Assumptions:
    Earth Atmospheric composition
    
    Inputs:
    power              [W]
    battery.
      mass_gain_factor [kg/W]
      
    Outputs:
      mdot             [kg/s]
    """
    
    #weight gain of battery (positive means mass loss)
    mdot = -(power) *(battery.mass_gain_factor)  
                
    return mdot


# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/AVL_Objects/Aircraft.py
#
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from RCAIDE.Framework.Core import  Data 

# package imports 
from .Wing import Wing
from .Body import Body


# ----------------------------------------------------------------------------------------------------------------------
# Aircraft
# ---------------------------------------------------------------------------------------------------------------------- 
class Aircraft(Data):
    """A data class defining the entire AVL aircraft geometry

    Assumptions:
        None

    Source:
        None

    Inputs:
        None

    Outputs:
        None

    Properties Used:
        N/A
    """    	

    def __defaults__(self):
        """ Defines the data structure and defaults of aircraft classes

        Assumptions:
            None

        Source:
            None

        Inputs:
            None

        Outputs:
            None

        Properties Used:
            N/A
        """ 		
        self.tag    = 'aircraft'
        self.wings  = Data()
        self.bodies = Data()

    def append_wing(self,wing):
        """ Appends wing geometry onto aircraft class

        Assumptions:
            None

        Source:
            None

        Inputs:
            None

        Outputs:
            None

        Properties Used:
            N/A
        """ 		
        # assert database type
        if not isinstance(wing,Wing):
            raise Exception('input component must be of type AVL.Data.Wing()')

        # store data
        self.wings.append(wing)
        return


    def append_body(self,body):
        """ Appends body geometry onto aircraft class

        Assumptions:
            None

        Source:
            None

        Inputs:
            None

        Outputs:
            None

        Properties Used:
            N/A
        """ 		
        # assert database type
        if not isinstance(body,Body):
            raise Exception('input component must be of type AVL.Data.Body()')

        # store data
        self.bodies.append(body)
        return




# RCAIDE/Library/Methods/Propulsors/Converters/compressor/append_compressor_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_compressor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_compressor_conditions(compressor,segment,propulsor_conditions): 
    propulsor_conditions[compressor.tag]         = Conditions()
    propulsor_conditions[compressor.tag].inputs  = Conditions()
    propulsor_conditions[compressor.tag].outputs = Conditions()
    return 


# RCAIDE/Library/Methods/Propulsors/Converters/Compression_Nozzle/__init__.py
# (c) Copyright 2023 Aerospace Research Community LLC

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
        
from .append_supersonic_nozzle_conditions   import append_supersonic_nozzle_conditions
from .compute_limited_geometry              import compute_limited_geometry
from .compute_supersonic_nozzle_performance import compute_supersonic_nozzle_performance


## @ingroup Methods-Propulsion-Rotor_Wake-Fidelity_One
# compute_wake_induced_velocity.py
# 
# Created:  Sep 2020, M. Clarke 
# Modified: Dec 2021, R. Erhard

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# package imports
import numpy as np 

## @ingroup Methods-Propulsion-Rotor_Wake-Fidelity_One
def compute_wake_induced_velocity(WD,VD,cpts,azi_start_idx=0,sigma=0.11,suppress_root=False):  
    """ This computes the velocity induced by the Fidelity One semi-prescribed vortex wake (PVW)
    on lifting surface control points

    Assumptions:  
    
    Source:   
    
    Inputs: 
    WD     - helical wake distribution points               [Unitless] 
    VD     - vortex distribution points on lifting surfaces [Unitless] 
    cpts   - control points in segment                      [Unitless] 

    Properties Used:
    N/A
    """    
    
    # control point, time step , blade number , location on blade 
    num_vortex_pts = len(WD.XA1[0,0,:])    # number of vortex points
    num_eval_pts   = VD.n_cp               # number of evaluation points
    
    dtype = np.float64

    # expand vortex points
    WXA1  = np.tile(WD.XA1.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    WYA1  = np.tile(WD.YA1.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))     
    WZA1  = np.tile(WD.ZA1.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    WXA2  = np.tile(WD.XA2.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    WYA2  = np.tile(WD.YA2.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    WZA2  = np.tile(WD.ZA2.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
                
    WXB1  = np.tile(WD.XB1.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    WYB1  = np.tile(WD.YB1.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    WZB1  = np.tile(WD.ZB1.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    WXB2  = np.tile(WD.XB2.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    WYB2  = np.tile(WD.YB2.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    WZB2  = np.tile(WD.ZB2.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    GAMMA = np.tile(WD.GAMMA.astype(dtype)[azi_start_idx,:,:,None], (1,1,num_eval_pts))
    
    # expand evaluation points
    XC    = np.tile(VD.XC.astype(dtype)[None,None,:],(cpts,num_vortex_pts,1))
    YC    = np.tile(VD.YC.astype(dtype)[None,None,:],(cpts,num_vortex_pts,1))
    ZC    = np.tile(VD.ZC.astype(dtype)[None,None,:],(cpts,num_vortex_pts,1))
    
    # -------------------------------------------------------------------------------------------
    # Compute velocity induced by horseshoe vortex segments on every control point by every panel
    # -------------------------------------------------------------------------------------------     
    # Create empty data structure
    V_ind = np.zeros((cpts,VD.n_cp,3))
     
    # compute influence of bound vortices 
    _ , res_C_AB = vortex(XC, YC, ZC, WXA1, WYA1, WZA1, WXB1, WYB1, WZB1,sigma,GAMMA,bv=True,WD=WD) 
    C_AB         = res_C_AB.transpose(1,3,0,2) 
    
    # compute influence of right vortex segment
    _ , res_C_BC = vortex(XC, YC, ZC, WXB1, WYB1, WZB1, WXB2, WYB2, WZB2,sigma,GAMMA)
    C_BC         = res_C_BC.transpose(1,3,0,2) 
    
    # compute influence of bottom vortex segment
    _ , res_C_CD = vortex(XC, YC, ZC, WXB2, WYB2, WZB2, WXA2, WYA2, WZA2,sigma,GAMMA) 
    C_CD         = res_C_CD.transpose(1,3,0,2) 
    
    # compute influence of left vortex segment 
    _ , res_C_DA = vortex(XC, YC, ZC, WXA2, WYA2, WZA2, WXA1, WYA1, WZA1,sigma,GAMMA) 
    C_DA         = res_C_DA.transpose(1,3,0,2) 
    
    # Add all the influences together
    V_ind =  row_reduction_summation(C_AB) + row_reduction_summation(C_BC)  + row_reduction_summation(C_CD) + row_reduction_summation(C_DA)   
    

    return V_ind
  
  
# -------------------------------------------------------------------------------
# vortex strength computation
# -------------------------------------------------------------------------------
## @ingroup Methods-Propulsion-Rotor_Wake-Fidelity_One
def vortex(X,Y,Z,X1,Y1,Z1,X2,Y2,Z2,sigma, GAMMA = 1, bv=False,WD=None,use_regularization_kernal=True):
    """ This computes the velocity induced on a control point by a segment
    of a horseshoe vortex that points from point 1 to point 2 for a filament with
    positive vortex strength.
    
    Assumptions:  
    None 
    
    Source: 
    Low-Speed Aerodynamics, Second Edition by Joseph katz, Allen Plotkin
    Pgs. 584(Literature), 579-586 (Fortran Code implementation)
    
    Inputs:
    GAMMA       - propeller/rotor circulation
    [X,Y,Z]     - location of control point  
    [X1,Y1,Z1]  - location of point 1 
    [X2,Y2,Z2]  - location of point 2
    Properties Used:
    N/A
    
    """   
    X_X1  = X-X1
    X_X2  = X-X2
    X2_X1 = X2-X1

    Y_Y1  = Y-Y1
    Y_Y2  = Y-Y2
    Y2_Y1 = Y2-Y1

    Z_Z1  = Z-Z1
    Z_Z2  = Z-Z2 
    Z2_Z1 = Z2-Z1 

    R1R2X  = Y_Y1*Z_Z2 - Z_Z1*Y_Y2 
    R1R2Y  = Z_Z1*X_X2 - X_X1*Z_Z2
    R1R2Z  = X_X1*Y_Y2 - Y_Y1*X_X2
    
    SQUARE = np.square(R1R2X) + np.square(R1R2Y) + np.square(R1R2Z)
    SQUARE[SQUARE==0] = 1e-8
    R1     = np.sqrt(np.square(X_X1) + np.square(Y_Y1) + np.square(Z_Z1)) 
    R2     = np.sqrt(np.square(X_X2) + np.square(Y_Y2) + np.square(Z_Z2)) 
    R0R1   = X2_X1*X_X1 + Y2_Y1*Y_Y1 + Z2_Z1*Z_Z1
    R0R2   = X2_X1*X_X2 + Y2_Y1*Y_Y2 + Z2_Z1*Z_Z2
    RVEC   = np.array([R1R2X,R1R2Y,R1R2Z])
    COEF   = (1/(4*np.pi))*(RVEC/SQUARE) * (R0R1/R1 - R0R2/R2)    

    
    if use_regularization_kernal:
        COEF = regularization_kernel(COEF, sigma)

    if bv:
        # ignore the row of panels corresponding to the lifting line of the rotor
        COEF_new = np.reshape(COEF[0,:,:,0],np.shape(WD.reshaped_wake.XA1[0,:,:,:,:]))
        m = np.shape(WD.reshaped_wake.XA1)[1]
        
        lifting_line_panels = np.zeros_like(COEF_new,dtype=bool)
        lifting_line_panels[:,:,:,0] = True
        lifting_line_panels_compressed = np.reshape(lifting_line_panels, (m,np.size(lifting_line_panels[0,:,:,:])))
        
        COEF[:,lifting_line_panels_compressed,:] = 0
    

    V_IND  = GAMMA * COEF
    
    return COEF , V_IND  

def row_reduction_summation(A):
    # sum along last axis
    sum_res = A.dot(np.ones(A.shape[-1])) # sum along axis
    
    return sum_res

## @ingroup Methods-Propulsion-Rotor_Wake-Fidelity_One
def regularization_kernel(COEF_in, sigma):
    """
    Regularization kernel used to prevent singularities
    
    Assumptions
    Spreads the vortex core over the radius sigma
    
    Source
       Winckelmans, "Topics in Vortex Methods for the Computation of Three-and Two-dimensional 
       Incompressible Unsteady Flows", 1989.
    
    Inputs:
       COEF    Biot-Savart Kernel
       sigma   regularization radius
    
    Outputs:
       KAPPA   Regularization Kernel
       
    Properties Used:
    N/A
    
    """
    COEF       = COEF_in.astype(np.float32)
    COEF_MAG   = np.abs(COEF)
    
    # Make sure the magnitude doesn't go to zero
    COEF_MAG[np.isclose(COEF_MAG,0,atol=1e-8)] = 1e-8

    R_square   = 1/(4*np.pi*COEF_MAG)
    R          = np.sqrt(R_square)
    R_sigma_sq = R_square/np.square(sigma)
    
    NUM = R*(R_sigma_sq + (5/2))
    DEN = 4*np.pi*(sigma**3)*np.power(R_sigma_sq + 1,5/2)
    
    DEN[DEN==0.] = 1e-8
    
    KAPPA = (NUM / DEN) * np.sign(COEF)
    
    return KAPPA


## @ingroup Methods-Propulsion
# fm_id.py
#
# Created:  ### ####, RCAIDE Team
# Modified: Feb 2016, E. Botero
#           Dec 2017, W. Maier

# ----------------------------------------------------------------------
#  fm_id
# ----------------------------------------------------------------------

## @ingroup Methods-Propulsion

def fm_id(M,gamma):
    """Function that takes in the Mach number and isentropic expansion factor,
    and outputs a value for f(M) that's commonly used in compressible flow
    calculations.

    Inputs:
    M       [-]
    gamma   [-]

    Outputs:
    fm      [-]

    Spurce:
    https://web.stanford.edu/~cantwell/AA210A_Course_Material/AA210A_Course_Notes/
    """

    m0 = (gamma+1.)/(2.*(gamma-1.))
    m1 = ((gamma+1.)/2.)**m0
    m2 = (1.+(gamma-1.)/2.*M*M)**m0
    fm = m1*M/m2

    return fm



# RCAIDE/Methods/Energy/Sources/Battery/Lithium_Ion_LFP/compute_lfp_cell_performance.py
# 
# 
# Created: Nov 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core import Units
import numpy as np  
from copy import deepcopy

# ----------------------------------------------------------------------------------------------------------------------
# compute_lfp_cell_performance
# ----------------------------------------------------------------------------------------------------------------------  
def compute_lfp_cell_performance(battery_module,state,bus,coolant_lines,t_idx, delta_t): 
    """
       Assumptions: 
        - All battery_module modules exhibit the same thermal behavior.
        - The cell temperature is assumed to be the temperature of the entire module.
       
       Source:
       Internal Resistance:
       
       Inputs:
         battery. 
                *** will be done in sphinx format.
           
         inputs.
        
       
       Outputs:
         battery.          
   
        
    """ 
    # ---------------------------------------------------------------------------------    
    # battery cell properties
    # --------------------------------------------------------------------------------- 
    electrode_area            = battery_module.cell.electrode_area 
    As_cell                   = battery_module.cell.surface_area
    cell_mass                 = battery_module.cell.mass    
    Cp                        = battery_module.cell.specific_heat_capacity       
    battery_module_data       = battery_module.cell.discharge_performance_map
    
    # ---------------------------------------------------------------------------------
    # Compute Bus electrical properties 
    # ---------------------------------------------------------------------------------    
    bus_conditions              = state.conditions.energy[bus.tag]
    bus_config                  = bus.battery_module_electric_configuration
    E_bus                       = bus_conditions.energy
    P_bus                       = bus_conditions.power_draw
    I_bus                       = bus_conditions.current_draw
    
    # ---------------------------------------------------------------------------------
    # Compute battery_module Conditions
    # -------------------------------------------------------------------------    
    battery_module_conditions = state.conditions.energy[bus.tag].battery_modules[battery_module.tag]  
   
    E_module_max       = battery_module.maximum_energy * battery_module_conditions.cell.capacity_fade_factor
    
    V_oc_module        = battery_module_conditions.voltage_open_circuit
    V_oc_cell          = battery_module_conditions.cell.voltage_open_circuit   
  
    P_module           = battery_module_conditions.power
    P_cell             = battery_module_conditions.cell.power
    
    
    Q_heat_module      = battery_module_conditions.heat_energy_generated
    Q_heat_cell        = battery_module_conditions.cell.heat_energy_generated
    
    V_ul_module        = battery_module_conditions.voltage_under_load
    V_ul_cell          = battery_module_conditions.cell.voltage_under_load
    
    I_module           = battery_module_conditions.current 
    I_cell             = battery_module_conditions.cell.current
    
    T_module           = battery_module_conditions.temperature                 
    T_cell             = battery_module_conditions.cell.temperature
    
    SOC_cell           = battery_module_conditions.cell.state_of_charge  
    SOC_module         = battery_module_conditions.state_of_charge
    E_cell             = battery_module_conditions.cell.energy   
    E_module           = battery_module_conditions.energy
    Q_cell             = battery_module_conditions.cell.charge_throughput              
    DOD_cell           = battery_module_conditions.cell.depth_of_discharge
    
    # ---------------------------------------------------------------------------------
    # Compute battery_module electrical properties 
    # -------------------------------------------------------------------------    
    # Calculate the current going into one cell  
    n_series          = battery_module.electrical_configuration.series
    n_parallel        = battery_module.electrical_configuration.parallel 
    n_total           = n_series * n_parallel
    no_modules        = bus.number_of_battery_modules
    
    # ---------------------------------------------------------------------------------
    # Examine Thermal Management System
    # ---------------------------------------------------------------------------------
    HAS = None  
    for coolant_line in coolant_lines:
        for tag, item in  coolant_line.items():
            if tag == 'battery_modules':
                for sub_tag, sub_item in item.items():
                    if sub_tag == battery_module.tag:
                        for btms in  sub_item:
                            HAS = btms    


    # ---------------------------------------------------------------------------------------------------
    # Current State 
    # ---------------------------------------------------------------------------------------------------
    if bus_config == 'Series':
        I_module[t_idx]      = I_bus[t_idx]
    elif bus_config  == 'Parallel':
        I_module[t_idx]      = I_bus[t_idx] / bus.number_of_battery_modules

    I_cell[t_idx] = I_module[t_idx] / n_parallel   
       
    # ---------------------------------------------------------------------------------
    # Compute battery_module cell temperature 
    # ---------------------------------------------------------------------------------
    # Determine temperature increase         
    sigma                 =  130  
    i_cell                = I_cell[t_idx]/electrode_area # current intensity (A/m²)
    q_dot_entropy         = (4.6810 * SOC_cell[t_idx]**4 + (-8.3729) * SOC_cell[t_idx]**3 + 3.7197 * SOC_cell[t_idx]**2 + 0.4356 * SOC_cell[t_idx]+ (-0.3027)) # Obtained from curve fitting the dUdt curve  
    q_dot_joule           = (i_cell**2)/(sigma)          
    Q_heat_cell[t_idx]    = (q_dot_joule + q_dot_entropy)*As_cell 
    Q_heat_module[t_idx]  = Q_heat_cell[t_idx]*n_total  
    V_ul_cell[t_idx]      = compute_lfp_cell_state(battery_module,battery_module_data,SOC_cell[t_idx],T_cell[t_idx],abs(I_cell[t_idx])) 
 
    # Effective Power flowing through battery_module 
    P_module[t_idx]       = P_bus[t_idx] /no_modules  - np.abs(Q_heat_module[t_idx]) 

    # store remaining variables  
    V_ul_module[t_idx]     = V_ul_cell[t_idx]*n_series  
    T_module[t_idx]        = T_cell[t_idx]   # Assume the cell temperature is the temperature of the module
    P_cell[t_idx]          = P_module[t_idx]/n_total 
    E_module[t_idx]        = E_bus[t_idx]/no_modules 
    E_cell[t_idx]          = E_module[t_idx]/n_total  
    # ---------------------------------------------------------------------------------------------------     
    # Future State 
    # --------------------------------------------------------------------------------------------------- 
    if t_idx != state.numerics.number_of_control_points-1:  

       # Compute cell temperature
        if HAS is not None:
            T_cell[t_idx+1]  = HAS.compute_thermal_performance(battery_module,bus,coolant_line,Q_heat_cell[t_idx],T_cell[t_idx],state,delta_t[t_idx],t_idx)
        else:
            # Considers a thermally insulated system and the heat piles on in the system
            dT_dt              = Q_heat_cell[t_idx]/(cell_mass*Cp)
            T_cell[t_idx+1]    = T_cell[t_idx] + dT_dt*delta_t[t_idx]
            
        # Compute state of charge and depth of discarge of the battery_module
        E_module[t_idx+1]                                     = np.float32(E_module[t_idx] -P_module[t_idx]*delta_t[t_idx])
        E_module[t_idx+1][E_module[t_idx+1] > E_module_max]   = np.float32(E_module_max)
        SOC_cell[t_idx+1]                                     = E_module[t_idx+1]/E_module_max 
        SOC_cell[t_idx+1][SOC_cell[t_idx+1]>1]                = 1.
        SOC_cell[t_idx+1][SOC_cell[t_idx+1]<0]                = 0. 
        DOD_cell[t_idx+1]                                     = 1 - SOC_cell[t_idx+1]  
        SOC_module[t_idx+1]                                   = SOC_cell[t_idx+1]
    
        # Determine new charge throughput (the amount of charge gone through the battery)
        Q_cell[t_idx+1]    = Q_cell[t_idx] + abs(I_cell[t_idx])*delta_t[t_idx]/Units.hr
        
    stored_results_flag     = True
    stored_battery_tag     = battery_module.tag  
        
    return stored_results_flag, stored_battery_tag
def reuse_stored_lfp_cell_data(battery_module,state,bus,coolant_lines, t_idx, delta_t,stored_results_flag, stored_battery_tag):
    '''Reuses results from one propulsor for identical batteries
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    

    Outputs:  
    
    Properties Used: 
    N.A.        
    '''
   
    state.conditions.energy[bus.tag].battery_modules[battery_module.tag] = deepcopy(state.conditions.energy[bus.tag].battery_modules[stored_battery_tag])      
    return


def compute_lfp_cell_state(battery_module,battery_module_data,SOC,T,I):
    """This computes the electrical state variables of a lithium ion 
    battery_module cell with a  lithium-nickel-cobalt-aluminum oxide cathode 
    chemistry from look-up tables 
     
    Assumtions: 
    N/A
    
    Source:  
    N/A 
     
    Inputs:
        SOC           - state of charge of cell     [unitless]
        battery_module_data  - look-up data structure      [unitless]
        T             - battery_module cell temperature    [Kelvin]
        I             - battery_module cell current        [Amperes]
    
    Outputs:  
        V_ul          - under-load voltage          [Volts] 
        
    """ 

    # Make sure things do not break by limiting current, temperature and current 
    capacity      = battery_module.cell.nominal_capacity
    SOC[SOC < 0.]   = 0.  
    SOC[SOC > 1.]   = 1.    
    DOD             = 1 - SOC 
    discharge_capacity = DOD*capacity
    

    T              = T-273
    # Operating Limits of the cell
    T[T<-10]       = -10 # model does not fit for below -10  degrees
    T[T>60]        =  60 # model does not fit for above 60 degrees

    
    I[I<0.0]      = 0.0
    I[I>52.0]     = 52.0
    C_rate        = I/capacity
     

    V_ul  = battery_module_data(C_rate, T, discharge_capacity)
    
    return V_ul


# RCAIDE/Library/Methods/Thermal_Management/Reservoirs/Reservoir_Tank/compute_mixing_temperature.py


# Created:  Apr 2024, S. Shekar 

# ---------------------------------------------------------------------------------------------------------------------- 
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
import  RCAIDE
import numpy as np
from scipy.optimize import fsolve


def compute_mixing_temperature(reservoir, state, coolant_line, delta_t, t_idx):
    """
    Computes the resultant temperature of the reservoir at each time step with coolant exchanging heat to the environment.

    :param reservoir: Reservoir Data Structure
        - reservoir.surface_area
        - reservoir.volume
        - reservoir.thickness
        - reservoir.material.conductivity
        - reservoir.material.emissivity
        - reservoir.coolant
    :type reservoir: dict
    :param state: State Data Structure
        - state.conditions.freestream.temperature
        - state.conditions.energy.coolant_line[reservoir.tag].coolant_temperature
    :type state: dict
    :param coolant_line: Coolant Line Data Structure
    :type coolant_line: dict
    :param delta_t: Time step
    :type delta_t: float
    :param t_idx: Time index
    :type t_idx: int
    :return: Updated temperature of the reservoir coolant
    :rtype: float

    :Assumptions: 
        N/A

    :Source:
        None
    """  
    T_current = 0
    volume    = 0
    for reservoir in coolant_line.reservoirs:
        T_current += state.conditions.energy[coolant_line.tag][reservoir.tag].coolant_temperature[t_idx, 0]
        volume    += reservoir.volume
    
    T_current = T_current / len(coolant_line.reservoirs)

    # Reservoir Properties
    coolant = reservoir.coolant
    rho_coolant = coolant.compute_density(T_current)
    Cp_RES = coolant.compute_cp(T_current)
    mass_coolant  = rho_coolant * volume

    mass_flow_HAS  = []
    T_outlet_HAS   = []
    Cp_HAS         = []
    mass_flow_HEX  = []
    T_outlet_HEX   = []
    Cp_HEX         = []

    for battery in coolant_line.battery_modules:
        for HAS in battery:
            if isinstance(HAS, RCAIDE.Library.Components.Thermal_Management.Batteries.Liquid_Cooled_Wavy_Channel):
                mass_flow_HAS.append(state.conditions.energy[coolant_line.tag][HAS.tag].coolant_mass_flow_rate[t_idx + 1])
                T_outlet_HAS.append(state.conditions.energy[coolant_line.tag][HAS.tag].outlet_coolant_temperature[t_idx + 1])
                Cp_HAS.append(coolant.compute_cp(T_outlet_HAS[-1]))

    for HEX in coolant_line.heat_exchangers:
        mass_flow_HEX.append(state.conditions.energy[coolant_line.tag][HEX.tag].coolant_mass_flow_rate[t_idx + 1])
        T_outlet_HEX.append(state.conditions.energy[coolant_line.tag][HEX.tag].outlet_coolant_temperature[t_idx + 1])
        Cp_HEX.append(coolant.compute_cp(T_outlet_HEX[-1]))

    # Solve for T_final using fsolve
        T_final = fsolve(energy_balance, T_current, args=(T_current, delta_t, mass_coolant, Cp_RES, Cp_HAS, Cp_HEX, mass_flow_HAS, T_outlet_HAS, mass_flow_HEX, T_outlet_HEX, reservoir, state, t_idx))[0]

    # Update the reservoir temperature
    state.conditions.energy[coolant_line.tag][reservoir.tag].coolant_temperature[t_idx + 1, 0] = T_final
    return

def compute_heat_loss_to_environment(T_final, T_ambient, reservoir):
    # Properties of Reservoir
    A_surface       = reservoir.surface_area
    thickness       = reservoir.thickness
    conductivity    = reservoir.material.conductivity
    emissivity_res  = reservoir.material.emissivity

    # Heat Transfer properties
    sigma           = 5.69e-8  # Stefan Boltzmann Constant
    h               = 1000  # [W/m^2-K]
    emissivity_air  = 0.9

    # Heat Transfer due to conduction
    dQ_dt_cond = conductivity * A_surface * (T_final - T_ambient) / thickness

    # Heat Transfer due to natural convection
    dQ_dt_conv = h * A_surface * (T_final - T_ambient)

    # Heat Transfer due to radiation
    dQ_dt_rad = sigma * A_surface * ((emissivity_res * T_final ** 4) - (emissivity_air * T_ambient ** 4))

    return dQ_dt_cond + dQ_dt_conv + dQ_dt_rad

def energy_balance(T_final, T_current, delta_t, mass_coolant, Cp_RES, Cp_HAS, Cp_HEX, mass_flow_HAS, T_outlet_HAS, mass_flow_HEX, T_outlet_HEX, reservoir, state, t_idx):
    # Ambient Air Temperature
    T_ambient = state.conditions.freestream.temperature[t_idx, 0]

    # Compute heat loss to the environment
    dQ_dt_env = compute_heat_loss_to_environment(T_final, T_ambient, reservoir)

    return (T_final - T_current
            - (delta_t / (mass_coolant * Cp_RES)) *
            (sum(mass_flow_HAS) * np.average(Cp_HAS) * (np.average(T_outlet_HAS) - T_final) +
             sum(mass_flow_HEX) * np.average(Cp_HEX) * (np.average(T_outlet_HEX) - T_final) -
             dQ_dt_env))


## @ingroup Optimization-Package_Setups
# particle_swarm_optimization.py
# 
# Created:  Sep. 2019, M. Clarke

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# RCAIDE imports
import numpy as np
import scipy as sp
 
## @ingroup Optimization-Package_Setups
def particle_swarm_optimization(func, lb, ub, ieqcons=[], f_ieqcons=None, args=(), kwargs={}, 
        swarmsize=100, omega=0.5, phip=0.5, phig=0.5, maxiter=100, 
        minstep=1e-8, minfunc=1e-8, debug=False):
    """
    This function perform a particle swarm optimization (PSO)
    
    Source:
        Pyswarm: https://github.com/tisimst/pyswarm
          
    Inputs: 
        func      : The function to be minimized                                                                    [function] 
        lb        : The lower bounds of the design variable(s)                                                      [array] 
        ub        :  The upper bounds of the design variable(s)                                                     [array]
                                                                                                                    
        ieqcons   : A list of functions of length n such that ieqcons[j](x,*args)                                   
                    >= 0.0 in a successfully optimized problem (Default: [])                                        [list]
        f_ieqcons : Returns a 1-D array in which each element must be greater or equal                              
                    to 0.0 in a successfully optimized problem. If f_ieqcons is specified,                          
                    ieqcons is ignored (Default: None)                                                              [function]
        args      : Additional arguments passed to objective and constraint functions                               [tuple]                       
        kwargs    : Additional keyword arguments passed to objective and constraint functions                       [dict] 
        swarmsize : The number of particles in the swarm (Default: 100)                                             [int] 
        omega     : Particle velocity scaling factor (Default: 0.5)                                                 [float] 
        phip      : Scaling factor to search away from the particle's best known position (Default: 0.5)            [scalar] 
        phig      : Scaling factor to search away from the swarm's best known position (Default: 0.5)               [scalar]
        maxiter   : The maximum number of iterations for the swarm to search (Default: 100) [int]                   
        minstep   : The minimum stepsize of swarm's best position before the search terminates (Default: 1e-8)      [scalar]
        minfunc   : The minimum change of swarm's best objective value before the search terminates (Default: 1e-8) [scalar]
        debug     : If True, progress statements will be displayed every iteration (Default: False)                 [boolean]
   
    Outputs:
        g         : The swarm's best known position (optimal design)                                                [list] 
        f         : The objective value at ``g``                                                                    [float]
         
    Properties Used:
        None
   
    """
    
    # License 
    ''' 
    Copyright (c) 2015, Sebastian M. Castillo Hair
    All rights reserved.
    
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:
    1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.
    3. All advertising materials mentioning features or use of this software
        must display the following acknowledgement:
        This product includes software developed by Sebastian M. Castillo Hair.
    4. Neither the name of Sebastian M. Castillo Hair nor the
        names of its contributors may be used to endorse or promote products
        derived from this software without specific prior written permission.
    
    THIS SOFTWARE IS PROVIDED BY SEBASTIAN M. CASTILLO HAIR ''AS IS'' AND ANY
    EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL SEBASTIAN M. CASTILLO HAIR BE LIABLE FOR ANY
    DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
    ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.'''
    
    assert len(lb)==len(ub), 'Lower- and upper-bounds must be the same length'
    assert hasattr(func, '__call__'), 'Invalid function handle'
    lb = np.array(lb)
    ub = np.array(ub)
    assert np.all(ub>lb), 'All upper-bound values must be greater than lower-bound values'
   
    vhigh = np.abs(ub - lb)
    vlow = -vhigh
    
    # Check for constraint function(s) #########################################
    obj = lambda x: func(x, *args, **kwargs)
    if f_ieqcons is None:
        if not len(ieqcons):
            if debug:
                print('No constraints given.')
            cons = lambda x: np.array([0])
        else:
            if debug:
                print('Converting ieqcons to a single constraint function')
            cons = lambda x: np.array([y(x, *args, **kwargs) for y in ieqcons])
    else:
        if debug:
            print('Single constraint function given in f_ieqcons')
        cons = lambda x: np.array(f_ieqcons(x, *args, **kwargs))
        
    def is_feasible(x):
        check = np.all(cons(x)>=0)
        return check
        
    # Initialize the particle swarm ############################################
    S = swarmsize
    D = len(lb)  # the number of dimensions each particle has
    x = np.random.rand(S, D)  # particle positions
    v = np.zeros_like(x)  # particle velocities
    p = np.zeros_like(x)  # best particle positions
    fp = np.zeros(S)  # best particle function values
    g = []  # best swarm position
    fg = 1e100  # artificial best swarm position starting value
    
    for i in range(S):
        # Initialize the particle's position
        x[i, :] = lb + x[i, :]*(ub - lb)
   
        # Initialize the particle's best known position
        p[i, :] = x[i, :]
       
        # Calculate the objective's value at the current particle's
        fp[i] = obj(p[i, :])
       
        # At the start, there may not be any feasible starting point, so just
        # give it a temporary "best" point since it's likely to change
        if i==0:
            g = p[0, :].copy()

        # If the current particle's position is better than the swarm's,
        # update the best swarm position
        if fp[i]<fg and is_feasible(p[i, :]):
            fg = fp[i]
            g = p[i, :].copy()
       
        # Initialize the particle's velocity
        v[i, :] = vlow + np.random.rand(D)*(vhigh - vlow)
       
    # Iterate until termination criterion met ##################################
    it = 1
    while it<=maxiter:
        rp = np.random.uniform(size=(S, D))
        rg = np.random.uniform(size=(S, D))
        for i in range(S):

            # Update the particle's velocity
            v[i, :] = omega*v[i, :] + phip*rp[i, :]*(p[i, :] - x[i, :]) + \
                      phig*rg[i, :]*(g - x[i, :])
                      
            # Update the particle's position, correcting lower and upper bound 
            # violations, then update the objective function value
            x[i, :] = x[i, :] + v[i, :]
            mark1 = x[i, :]<lb
            mark2 = x[i, :]>ub
            x[i, mark1] = lb[mark1]
            x[i, mark2] = ub[mark2]
            fx = obj(x[i, :])
            
            # Compare particle's best position (if constraints are satisfied)
            if fx<fp[i] and is_feasible(x[i, :]):
                p[i, :] = x[i, :].copy()
                fp[i] = fx

                # Compare swarm's best position to current particle's position
                # (Can only get here if constraints are satisfied)
                if fx<fg:
                    if debug:
                        print('New best for swarm at iteration {:}: {:} {:}'.format(it, x[i, :], fx))

                    tmp = x[i, :].copy()
                    stepsize = np.sqrt(np.sum((g-tmp)**2))
                    if np.abs(fg - fx)<=minfunc:
                        print('Stopping search: Swarm best objective change less than {:}'.format(minfunc))
                        return tmp, fx
                    elif stepsize<=minstep:
                        print('Stopping search: Swarm best position change less than {:}'.format(minstep))
                        return tmp, fx
                    else:
                        g = tmp.copy()
                        fg = fx

        if debug:
            print('Best after iteration {:}: {:} {:}'.format(it, g, fg))
        it += 1

    print('Stopping search: maximum iterations reached --> {:}'.format(maxiter))
    
    if not is_feasible(g):
        print("However, the optimization couldn't find a feasible design. Sorry")
    return g, fg




# RCAIDE/Framework/Core/Arrays.py

# ----------------------------------------------------------------------
#   Imports
# ----------------------------------------------------------------------
import numpy as np

# ----------------------------------------------------------------------
#   Array
# ----------------------------------------------------------------------       

array_type  = np.ndarray
matrix_type = np.matrixlib.defmatrix.matrix
 
def atleast_2d_col(A):
    """Makes a 2D array in column format

    Assumptions:
    None

    Source:
    N/A

    Inputs:
    A      [1-D Array]

    Outputs:
    A      [2-D Array]

    Properties Used:
    N/A
    """       
    return atleast_2d(A,'col')


def atleast_2d_row(A):
    """Makes a 2D array in row format

    Assumptions:
    None

    Source:
    N/A

    Inputs:
    A      [1-D Array]

    Outputs:
    A      [2-D Array]

    Properties Used:
    N/A
    """       
    return atleast_2d(A,'row')


def atleast_2d(A,oned_as='row'):
    """ensures A is an array and at least of rank 2

    Assumptions:
    Defaults as row

    Source:
    N/A

    Inputs:
    A      [1-D Array]

    Outputs:
    A      [2-D Array]

    Properties Used:
    N/A
    """       
    
    # not an array yet
    if not isinstance(A,(array_type,matrix_type)):
        if not isinstance(A,(list,tuple)):
            A = [A]
        A = np.array(A)
        
    # check rank
    if A.ndim < 2:
        # expand row or col
        if oned_as == 'row':
            A = A[None,:]
        elif oned_as == 'col':
            A = A[:,None]
        else:
            raise Exception("oned_as must be 'row' or 'col' ")
            
    return A


# RCAIDE/Library/Components/Propulsors/Converters/Lift_Rotor.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
# RCAIDE imports 
from RCAIDE.Framework.Core import Data
from .Rotor import Rotor

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  LIFT ROTOR CLASS
# ----------------------------------------------------------------------------------------------------------------------           
class Lift_Rotor(Rotor):
    """
    A lift rotor component model for vertical lift applications, inheriting from the base Rotor class.

    Attributes
    ----------
    tag : str
        Identifier for the lift rotor. Default is 'lift_rotor'.
        
    orientation_euler_angles : list
        Vector of angles [rad] defining rotor orientation [θ, φ, ψ]. 
        Default is [0., π/2, 0.] for Z-direction thrust up in vehicle frame.
        
    use_2d_analysis : bool
        Flag for using 2D aerodynamic analysis. Default is False.
        
    variable_pitch : bool
        Flag indicating if rotor has variable pitch capability. Default is False.
        
    hover : Data
        Hover performance parameters
        
        - design_thrust : float
            Design point thrust [N]. Default is None.
        
        - design_torque : float
            Design point torque [N·m]. Default is None.

        - design_power : float
            Design point power [W]. Default is None.
        
        - design_angular_velocity : float
            Design point rotational speed [rad/s]. Default is None.

        - design_tip_mach : float
            Design point blade tip Mach number. Default is None.
        
        - design_freestream_velocity : float
            Design point forward velocity [m/s]. Default is None.

        - design_acoustics : Data
            Acoustic characteristics at design point. Default is None.

        - design_performance : Data
            Performance metrics at design point. Default is None.

        - design_pitch_command : float
            Design point blade pitch command [rad]. Default is 0.0.

        - design_SPL_dBA : float
            Design point sound pressure level [dBA]. Default is None.

        - design_Cl : float
            Design point lift coefficient. Default is None.

        - design_thrust_coefficient : float
            Design point thrust coefficient. Default is None.

        - design_power_coefficient : float
            Design point power coefficient. Default is None.
            
    oei : Data
        One engine inoperative performance parameters
        
        - design_thrust : float
            OEI thrust requirement [N]. Default is None.

        - design_torque : float
            OEI torque requirement [N·m]. Default is None.

        - design_power : float
            OEI power requirement [W]. Default is None.

        - design_angular_velocity : float
            OEI rotational speed [rad/s]. Default is None.

        - design_freestream_velocity : float
            OEI forward velocity [m/s]. Default is None.

        - design_tip_mach : float
            OEI blade tip Mach number. Default is None.

        - design_altitude : float
            OEI design altitude [m]. Default is None.
        
        - design_acoustics : Data
            OEI acoustic characteristics. Default is None.

        - design_pitch_command : float
            OEI blade pitch command [rad]. Default is 0.0.

        - design_performance : Data
            OEI performance metrics. Default is None.

        - design_SPL_dBA : float
            OEI sound pressure level [dBA]. Default is None.

        - design_Cl : float
            OEI lift coefficient. Default is None.

        - design_thrust_coefficient : float
            OEI thrust coefficient. Default is None.

        - design_power_coefficient : float
            OEI power coefficient. Default is None.

    Notes
    -----
    The Lift_Rotor class models rotors specifically designed for vertical lift applications.
    It includes capabilities for:
    * Hover performance analysis
    * One engine inoperative (OEI) conditions
    * Variable pitch operation
    * Acoustic analysis
    * Performance coefficient calculations

    **Definitions**

    'OEI'
        One Engine Inoperative - emergency condition where one engine fails

    'SPL'
        Sound Pressure Level - measure of acoustic intensity
        
    'Tip Mach'
        Mach number at the blade tip, including rotational and forward flight effects

    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Converters.Rotor
    """     
    def __defaults__(self):
        """This sets the default values for the component to function.
        
        Assumptions:
        None
        
        Source:
        N/A
        
        Inputs:
        None
        
        Outputs:
        None
        
        Properties Used:
        None
        """         


        self.tag                              = 'lift_rotor'
        self.orientation_euler_angles         = [0.,np.pi/2.,0.] # This is Z-direction thrust up in vehicle frame
        self.use_2d_analysis                  = False
        self.variable_pitch                   = False 

        self.hover                            = Data()    
        self.hover.design_thrust              = None
        self.hover.design_torque              = None
        self.hover.design_power               = None
        self.hover.design_angular_velocity    = None
        self.hover.design_tip_mach            = None
        self.hover.design_freestream_velocity = None
        self.hover.design_acoustics           = None
        self.hover.design_performance         = None
        self.hover.design_pitch_command       = 0.0
        self.hover.design_SPL_dBA             = None
        self.hover.design_Cl                  = None
        self.hover.design_thrust_coefficient  = None
        self.hover.design_power_coefficient   = None 
        
        self.oei                              = Data()
        self.oei.design_thrust                = None
        self.oei.design_torque                = None
        self.oei.design_power                 = None
        self.oei.design_angular_velocity      = None
        self.oei.design_freestream_velocity   = None
        self.oei.design_tip_mach              = None  
        self.oei.design_altitude              = None
        self.oei.design_acoustics             = None
        self.oei.design_pitch_command         = 0.0
        self.oei.design_performance           = None
        self.oei.design_SPL_dBA               = None
        self.oei.design_Cl                    = None
        self.oei.design_thrust_coefficient    = None
        self.oei.design_power_coefficient     = None         

        


# RCAIDE/Framework/Analyses/Mission/Segments/Cruise/Constant_Pitch_Rate_Constant_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports  
from RCAIDE.Framework.Mission.Segments.Evaluate   import Evaluate 
from RCAIDE.Framework.Core                                 import Units   
from RCAIDE.Library.Mission.Segments             import Cruise
from RCAIDE.Library.Mission                      import Common,Segments

# ----------------------------------------------------------------------------------------------------------------------
#  Constant_Pitch_Rate_Constant_Altitude
# ----------------------------------------------------------------------------------------------------------------------  

class Constant_Pitch_Rate_Constant_Altitude(Evaluate):
    """ Vehicle flies at a constant pitch rate at a set altitude. This is maneuvering flight.
        This is used in VTOL aircraft which need to transition from one pitch attitude to another.
    
        Assumptions:
        None
        
        Source:
        None
    """      
    
    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        
        # -------------------------------------------------------------------------------------------------------------- 
        #   User Inputs
        # -------------------------------------------------------------------------------------------------------------- 
        self.altitude          = None
        self.pitch_rate        = 1.  * Units['rad/s/s']
        self.pitch_initial     = None
        self.pitch_final       = 0.0 * Units['rad']
        self.true_course       = 0.0 * Units.degrees  
 
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------       
        initialize                         = self.process.initialize  
        initialize.conditions              = Cruise.Constant_Pitch_Rate_Constant_Altitude.initialize_conditions  
        iterate                            = self.process.iterate 
        iterate.unknowns.mission           = Common.Unpack_Unknowns.orientation  
        iterate.unknowns.controls          = Common.Unpack_Unknowns.control_surfaces
        iterate.residuals.flight_dynamics  = Common.Residuals.flight_dynamics
        
        return




# RCAIDE/Library/Methods/Weights/Physics_Based_Buildups/Common
# Created:  Jun 2017, J. Smart
# Modified: Apr 2018, J. Smart
#           Mar 2020, M. Clarke

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------
from RCAIDE.Library.Attributes.Materials import  Bidirectional_Carbon_Fiber, Carbon_Fiber_Honeycomb, Paint, Unidirectional_Carbon_Fiber, Aluminum, Epoxy, Nickel, Aluminum_Rib 
import numpy as np
import copy as cp

#-------------------------------------------------------------------------------
# compute_rotor_weight
#-------------------------------------------------------------------------------
def compute_rotor_weight(rotor,
         maximum_lifting_thrust,
         chord_to_radius_ratio = 0.1,
         thickness_to_chord = 0.12,
         root_to_radius_ratio = 0.1,
         moment_to_lift_ratio = 0.02,
         spanwise_analysis_points = 5,
         safety_factor = 1.5,
         margin_factor = 1.2,
         forward_web_locations = [0.25, 0.35],
         shear_center = 0.25,
         speed_of_sound = 340.294,
         tip_max_mach_number = 0.65):
    """ Calculates propeller blade pass for an eVTOL vehicle based on assumption
        of a NACA airfoil rotor, an assumed cm/cl, tip Mach limit, and structural
        geometry. 

        Assumptions:
        Originally written as part of an AA 290 project inteded for trade study
        of the above vehicle types.

        If vehicle model does not have material properties assigned, appropriate
        assumptions are made based on RCAIDE's Solids Attributes library.
        
        Sources:
        Project Vahana Conceptual Trade Study

        Inputs:

            rotor                       RCAIDE Rotor Data Structure
            maximum_thrust              Maximum Design Thrust               [N]
            chord_to_radius_ratio       Chord to Blade Radius               [Unitless]
            thickness_to_chord          Blade Thickness to Chord            [Unitless]
            root_to_radius_ratio        Root Structure to Blade Radius      [Unitless]
            moment_to_lift_ratio        Coeff. of Moment to Coeff. of Lift  [Unitless]
            spanwise_analysis_points    Analysis Points for Sizing          [Unitless]
            safety_factor               Design Safety Factor                [Unitless]
            margin_factor               Allowable Extra Mass Fraction       [Unitless]
            forward_web_locationss      Location of Forward Spar Webbing    [m]
            shear_center                Location of Shear Center            [m]
            speed_of_sound              Local Speed of Sound                [m/s]
            tip_max_mach_number         Allowable Tip Mach Number           [Unitless]

        Outputs:

            weight:                 Propeller Mass                      [kg]
            
        Properties Used:
        Material properties of imported RCAIDE Solids
    """

    #-------------------------------------------------------------------------------
    # Unpack Inputs
    #-------------------------------------------------------------------------------
    rProp              = rotor.tip_radius
    maxLiftingThrust   = maximum_lifting_thrust
    nBlades            = rotor.number_of_blades
    chord              = rProp * chord_to_radius_ratio
    N                  = spanwise_analysis_points
    SF                 = safety_factor
    toc                = thickness_to_chord
    fwdWeb             = cp.deepcopy(forward_web_locations)
    xShear             = shear_center
    rootLength         = rProp * root_to_radius_ratio
    grace              = margin_factor
    sound              = speed_of_sound
    tipMach            = tip_max_mach_number
    cmocl              = moment_to_lift_ratio
    
    #-------------------------------------------------------------------------------
    # Unpack Material Properties
    #-------------------------------------------------------------------------------

    try:
        torsMat = rotor.materials.skin_materials.torsion_carrier
    except AttributeError:
        torsMat = Bidirectional_Carbon_Fiber()
    torsUSS = torsMat.ultimate_shear_strength
    torsMGT = torsMat.minimum_gage_thickness
    torsDen = torsMat.density

    try:
        shearMat = rotor.materials.spar_materials.shear_carrier
    except AttributeError:
        shearMat = Bidirectional_Carbon_Fiber()
    shearUSS = shearMat.ultimate_shear_strength
    shearMGT = shearMat.minimum_gage_thickness
    shearDen = shearMat.density

    try:
        bendMat = rotor.materials.flap_materials.bending_carrier
    except AttributeError:
        bendMat = Unidirectional_Carbon_Fiber()
    bendDen = bendMat.density
    bendUTS = bendMat.ultimate_tensile_strength
    bendUSS = bendMat.ultimate_shear_strength

    try:
        coreMat = rotor.materials.skin_materials.core
    except AttributeError:
        coreMat = Carbon_Fiber_Honeycomb()
    coreDen = coreMat.density

    try:
        ribMat = rotor.materials.rib_materials.structural
    except AttributeError:
        ribMat = Aluminum_Rib()
    ribWid = ribMat.minimum_width
    ribMGT = ribMat.minimum_gage_thickness
    ribDen = ribMat.density

    try:
        rootMat = rotor.materials.root_materials.structural
    except AttributeError:
        rootMat = Aluminum()
    rootDen = rootMat.density
    rootUTS = rootMat.ultimate_tensile_strength

    try:
        leMat = rotor.materials.skin_materials.leading_edge
    except:
        leMat = Nickel()
    leDen = leMat.density
    
    try:
        glueMat = rotor.materials.skin_materials.adhesive
    except AttributeError:
        glueMat = Epoxy()
    glueMGT = glueMat.minimum_gage_thickness
    glueDen = glueMat.density
    
    try:
        coverMat = rotor.materials.skin_materials.cover
    except:
        coverMat = Paint()
    coverMGT = coverMat.minimum_gage_thickness
    coverDen = coverMat.density

    #-------------------------------------------------------------------------------
    # Airfoil
    #-------------------------------------------------------------------------------
    NACA       = np.multiply(5 * toc, [0.2969, -0.1260, -0.3516, 0.2843, -0.1015])
    coord      = np.unique(fwdWeb+np.linspace(0,1,N).tolist())[:,np.newaxis]
    coordMAT   = np.concatenate((coord**0.5,coord,coord**2,coord**3,coord**4),axis=1)
    nacaMAT    = coordMAT.dot(NACA)[:, np.newaxis]
    coord      = np.concatenate((coord,nacaMAT),axis=1)
    coord      = np.concatenate((coord[-1:0:-1],coord.dot(np.array([[1.,0.],[0.,-1.]]))),axis=0)
    coord[:,0] = coord[:,0] - xShear

    #-------------------------------------------------------------------------------
    # Beam Geometry
    #-------------------------------------------------------------------------------
    x         = np.linspace(0,rProp,N)
    dx        = x[1] - x[0]
    fwdWeb[:] = [round(loc - xShear,2) for loc in fwdWeb]

    #-------------------------------------------------------------------------------
    # Loads
    #-------------------------------------------------------------------------------
    omega = sound*tipMach/rProp                              # Propeller Angular Velocity
    F     = SF*3*(maxLiftingThrust/rProp**3)*(x**2)/nBlades  # Force Distribution
    Q     = F * chord * cmocl                                # Torsion Distribution

    #-------------------------------------------------------------------------------
    # Initial Mass Estimates
    #-------------------------------------------------------------------------------
    box               = coord * chord
    skinLength        = np.sum(np.sqrt(np.sum(np.diff(box,axis=0)**2,axis=1)))
    maxThickness      = (np.amax(box[:,1])-np.amin(box[:,1]))/2
    rootBendingMoment = SF*maxLiftingThrust/nBlades*0.75*rProp
    m                 = (bendDen*dx*rootBendingMoment/
                        (2*bendUSS*maxThickness))+ \
                        skinLength*shearMGT*dx*shearDen
    m                 = m*np.ones(N)
    error             = 1               # Initialize Error
    tolerance         = 1e-8        # Mass Tolerance
    massOld           = np.sum(m)

    #-------------------------------------------------------------------------------
    # General Structural Properties
    #-------------------------------------------------------------------------------
    seg = []                                            # List of Structural Segments

    # Torsion
    enclosedArea = 0.5*np.abs(np.dot(box[:,0],np.roll(box[:,1],1))-
        np.dot(box[:,1],np.roll(box[:,0],1)))           # Shoelace Formula

    # Flap Properties
    box = coord                                         # Box Initially Matches Airfoil
    box = box[box[:,0]<=fwdWeb[1]]                      # Trim Coordinates Aft of Aft Web
    box = box[box[:,0]>=fwdWeb[0]]                      # Trim Coordinates Fwd of Fwd Web
    seg.append(box[box[:,1]>np.mean(box[:,1])]*chord)   # Upper Fwd Segment
    seg.append(box[box[:,1]<np.mean(box[:,1])]*chord)   # Lower Fwd Segment

    # Flap & Drag Inertia
    capInertia = 0
    capLength  = 0

    for i in range(0,2):
        l = np.sqrt(np.sum(np.diff(seg[i],axis=0)**2,axis=1))   # Segment Lengths
        c = (seg[i][1::]+seg[i][0::-1])/2                       # Segment Centroids

        capInertia += np.abs(np.sum(l*c[:,1] **2))
        capLength  += np.sum(l)

    # Shear Properties
    box = coord
    box = box[box[:,0]<=fwdWeb[1]]
    z   = box[box[:,0]==fwdWeb[0],1]*chord
    shearHeight = np.abs(z[0] - z[1])

    # Core Properties
    box = coord
    box = box[box[:,0]>=fwdWeb[0]]
    box = box*chord
    coreArea = 0.5*np.abs(np.dot(box[:,0],np.roll(box[:,1],1))-
        np.dot(box[:,1],np.roll(box[:,0],1)))                       # Shoelace Formula

    # Shear/Moment Calculations
    Vz = np.append(np.cumsum(( F[0:-1]*np.diff(x))[::-1])[::-1],0)  # Bending Moment
    Mx = np.append(np.cumsum((Vz[0:-1]*np.diff(x))[::-1])[::-1],0)  # Torsion Moment
    My = np.append(np.cumsum(( Q[0:-1]*np.diff(x))[::-1])[::-1],0)  # Drag Moment

    #-------------------------------------------------------------------------------
    # Mass Calculation
    #-------------------------------------------------------------------------------
    while error > tolerance:
        CF = (SF*omega**2*
            np.append(np.cumsum(( m[0:-1]*np.diff(x)*x[0:-1])[::-1])[::-1],0))  # Centripetal Force

        # Calculate Skin Weight Based on Torsion
        tTorsion = My/(2*torsUSS*enclosedArea)                 # Torsion Skin Thickness
        tTorsion = np.maximum(tTorsion,torsMGT*np.ones(N))     # Gage Constraint
        mTorsion = tTorsion * skinLength * torsDen             # Torsion Mass

        # Calculate Flap Mass Based on Bending
        tFlap = CF/(capLength*bendUTS) +   \
            Mx*np.amax(np.abs(box[:,1]))/(capInertia*bendUTS)
        mFlap = tFlap*capLength*bendDen
        mGlue = glueMGT*glueDen*capLength*np.ones(N)

        # Calculate Web Mass Based on Shear
        tShear = 1.5*Vz/(shearUSS*shearHeight)
        tShear = np.maximum(tShear,shearMGT*np.ones(N))
        mShear = tShear*shearHeight*shearDen

        # Paint Weight
        mPaint = skinLength*coverMGT*coverDen*np.ones(N)

        # Core Mass
        mCore = coreArea*coreDen*np.ones(N)
        mGlue += glueMGT*glueDen*skinLength*np.ones(N)

        # Leading Edge Protection
        box      = coord * chord
        box      = box[box[:,0]<(0.1*chord)]
        leLength = np.sum(np.sqrt(np.sum(np.diff(box,axis=0)**2,axis=1)))
        mLE      = leLength*420e-6*leDen*np.ones(N)

        # Section Mass
        m = mTorsion + mCore + mFlap + mShear + mGlue + mPaint + mLE

        # Rib Weight
        mRib = (enclosedArea+skinLength*ribWid)*ribMGT*ribDen

        # Root Fitting
        box   = coord * chord
        rRoot = (np.amax(box[:,1])-np.amin(box[:,1]))/2
        t     = np.amax(CF)/(2*np.pi*rRoot*rootUTS) +    \
                np.amax(Mx)/(3*np.pi*rRoot**2*rootUTS)
        mRoot = 2*np.pi*rRoot*t*rootLength*rootDen

        # Total Weight
        mass    = nBlades*(np.sum(m[0:-1]*np.diff(x))+2*mRib+mRoot)
        error   = np.abs(mass-massOld)
        massOld = mass

    mass = mass * grace

    return mass


# RCAIDE/Library/Attributes/Propellants/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Propellant           import Propellant
from .Aviation_Gasoline    import Aviation_Gasoline
from .Jet_A                import Jet_A
from .Jet_A1               import Jet_A1 
from .Gaseous_Hydrogen     import Gaseous_Hydrogen
from .Liquid_Natural_Gas   import Liquid_Natural_Gas
from .Liquid_Petroleum_Gas import Liquid_Petroleum_Gas
from .JP7                  import JP7
from .Methane              import Methane
from .Propane              import Propane
from .Ethane               import Ethane
from .Ethanol              import Ethanol
from .Liquid_Hydrogen      import Liquid_Hydrogen
from .Rocket_LH2           import Rocket_LH2
from .Rocket_RP1           import Rocket_RP1
from .Butanol		   import Butanol
from .Propanol		   import Propanol
from .Alkane_Mixture	   import Alkane_Mixture
from .Alcohol_Mixture	   import Alcohol_Mixture


# RCAIDE/Library/Compoments/Wings/Segment.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core     import Data, Container
from RCAIDE.Library.Components import Component   

# ---------------------------------------------------------------------------------------------------------------------- 
#  Segment
# ----------------------------------------------------------------------------------------------------------------------   
class Segment(Component):
    """
    A class representing a wing segment that defines a portion of a wing's geometry.

    Attributes
    ----------
    tag : str
        Unique identifier for the segment, defaults to 'segment'
        
    prev : Component
        Link to previous segment in wing, defaults to None
        
    next : Component
        Link to next segment in wing, defaults to None
        
    percent_span_location : float
        Spanwise location as fraction of wing span, defaults to 0.0
        
    twist : float
        Local twist angle, defaults to 0.0
        
    taper : float
        Local taper ratio, defaults to 0.0
        
    root_chord_percent : float
        Root chord as fraction of wing root chord, defaults to 0.0
        
    dihedral_outboard : float
        Dihedral angle of outboard section, defaults to 0.0
        
    thickness_to_chord : float
        Local thickness-to-chord ratio, defaults to 0.0
        
    sweeps : Data
        Collection of sweep angles
        
        - quarter_chord : float
            Quarter-chord sweep angle, defaults to 0.0
        - leading_edge : float
            Leading edge sweep angle, defaults to None
            
    chords : Data
        Collection of chord lengths
        
        - mean_aerodynamic : float
            Mean aerodynamic chord, defaults to 0.0
            
    areas : Data
        Collection of area measurements
        
        - reference : float
            Reference area, defaults to 0.0
        - exposed : float
            Exposed area, defaults to 0.0
        - wetted : float
            Wetted area, defaults to 0.0
            
    Airfoil : Container
        Collection of airfoil definitions, initialized empty

    Notes
    -----
    Wing segments are used to build up complete wing geometries by defining 
    local properties at various spanwise locations.

    See Also
    --------
    RCAIDE.Library.Components.Wings.Wing
        Parent wing class that uses segments
    RCAIDE.Library.Components.Wings.Main_Wing
        Main wing implementation using segments
    """ 

    def __defaults__(self):
        """
        Sets default values for the wing segment attributes.
        """         
        self.tag                     = 'segment'
        self.prev                    = None
        self.next                    = None  
        self.percent_span_location   = 0.0
        self.twist                   = 0.0
        self.taper                   = 0.0
        self.root_chord_percent      = 0.0
        self.dihedral_outboard       = 0.0
        self.thickness_to_chord      = 0.0
        
        self.sweeps                  = Data()
        self.sweeps.quarter_chord    = 0.0
        self.sweeps.leading_edge     = None
    
        self.chords                  = Data()
        self.chords.mean_aerodynamic = 0.0
        
        self.areas                   = Data()
        self.areas.reference         = 0.0
        self.areas.exposed           = 0.0
        self.areas.wetted            = 0.0

        self.airfoil                 = None
        
    def append_airfoil(self, airfoil):
        """
        Adds an airfoil definition to the segment.

        Parameters
        ----------
        airfoil : Data
            Airfoil data to be added to the segment
        """  
        # Assert database type
        if not isinstance(airfoil,RCAIDE.Library.Components.Airfoils.Airfoil):
            raise Exception('input component must be of type Airfoil')

        # store data
        self.airfoil = airfoil

         
class Segment_Container(Container):
    """
    Container class for managing wing segments. Provides organization and 
    access methods for segment components.

    See Also
    --------
    RCAIDE.Library.Components.Wings.Segments.Segment
        The segment components stored in this container
    """     

    def get_children(self):
        """
        Returns a list of allowable child component types for the segment container.

        Returns
        -------
        list
            Empty list as segments do not contain child components
        """       
        return []



# import_airfoil_dat.py
# 
# Created:  
# Modified: Sep 2016, E. Botero

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np

# ------------------------------------------------------------
#  import airfoil dat
# ------------------------------------------------------------ 
def import_airfoil_dat(filename):
    """Import an airfoil data file and stores it in a numpy array.
    
    Assumptions:
    Airfoil file in Lednicer format

    Source:
    None

    Inputs:
    filename   <string>

    Outputs:
    data       numpy array with airfoil data

    Properties Used:
    N/A
    """     
    
    filein = open(filename,'r')
    data = {}   
    data['header'] = filein.readline().strip() + filein.readline().strip()

    filein.readline()
    
    sections = ['upper','lower']
    data['upper'] = []
    data['lower'] = []
    section = None
    
    while True:
        line = filein.readline()
        if not line: 
            break
        
        line = line.strip()
        
        if line and section:
            pass
        elif not line and not section:
            continue
        elif line and not section:
            section = sections.pop(0)
        elif not line and section:
            section = None
            continue
        
        point = list(map(float,line.split()))
        data[section].append(point)
        
    for k,v in data.items():
        data[k] = np.array(v)
        
    return data


# RCAIDE/Library/Methods/Propulsors/Converters/Fan/compute_fan_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke    

# ---------------------------------------------------------------------------------------------------------------------- 
# Imports 
# ----------------------------------------------------------------------------------------------------------------------
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Fan 
# ----------------------------------------------------------------------------------------------------------------------            
def compute_fan_performance(fan,fan_conditions,conditions):
    """ This computes the output values from the input values according to
    equations from the source. The following outputs are computed: 
    fan_conditions.outputs.
      stagnation_temperature  (numpy.ndarray): exit stagnation_temperature  [K]  
      stagnation_pressure     (numpy.ndarray): exit stagnation_pressure     [Pa]
      stagnation_enthalpy     (numpy.ndarray): exit stagnation_enthalpy     [J/kg]
      work_done               (numpy.ndarray): work_done                    [J/(kg-s)]
 

    Assumptions:
        Constant polytropic efficiency and pressure ratio

    Source:
        https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/

    Args:
        conditions.freestream.
          isentropic_expansion_factor         (numpy.ndarray): isentropic_expansion_factor         [unitless]
          specific_heat_at_constant_pressure  (numpy.ndarray): specific_heat_at_constant_pressure  [J/(kg K)]
        fan
          .inputs.stagnation_temperature      (numpy.ndarray): entering stagnation temperature [K]
          .inputs.stagnation_pressure         (numpy.ndarray): entering stagnation pressure    [Pa] 
          .pressure_ratio                             (float): pressure ratio of fan           [unitless]
          .polytropic_efficiency                      (float): polytropic_efficiency           [unitless]
      
    Returns:
        None
        
    """        
     
    # unpack from fan
    PR                      = fan.pressure_ratio
    etapold                 = fan.polytropic_efficiency
    Tt_in                   = fan_conditions.inputs.stagnation_temperature
    Pt_in                   = fan_conditions.inputs.stagnation_pressure 
    P0                      = fan_conditions.inputs.static_pressure 
    T0                      = fan_conditions.inputs.static_temperature
    M0                      = fan_conditions.inputs.mach_number    
    
    # Unpack ram inputs
    working_fluid           = fan.working_fluid
 
    # Compute the working fluid properties 
    gamma  = working_fluid.compute_gamma(T0,P0) 
    Cp     = working_fluid.compute_cp(T0,P0)    
    
    # Compute the output quantities  
    Pt_out    = Pt_in*PR
    Tt_out    = Tt_in*PR**((gamma-1)/(gamma*etapold))
    T_out     = Tt_out/(1.+(gamma-1.)/2.*M0*M0)
    P_out     = Pt_out/((1.+(gamma-1.)/2.*M0*M0)**(gamma/(gamma-1.))) 
    ht_out    = Tt_out*Cp   
    ht_in     = Tt_in*Cp 
    M_out     = np.sqrt( (((Pt_out/P_out)**((gamma-1.)/gamma))-1.) *2./(gamma-1.) )     
    
    # Compute the work done by the fan (normalized by mass flow i.e. J/(kg/s)
    work_done = ht_out - ht_in
    
    # Store computed quantities into outputs
    fan_conditions.outputs.stagnation_temperature  = Tt_out
    fan_conditions.outputs.stagnation_pressure     = Pt_out
    fan_conditions.outputs.static_temperature      = T_out
    fan_conditions.outputs.static_pressure         = P_out    
    fan_conditions.outputs.work_done               = work_done
    fan_conditions.outputs.stagnation_enthalpy     = ht_out
    fan_conditions.outputs.mach_number             = M_out
    
    return 


# RCAIDE/Library/Compoments/Systems/System.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
# RCAIDE imports  
from RCAIDE.Library.Components import Component
 
# ----------------------------------------------------------------------------------------------------------------------
# System
# ----------------------------------------------------------------------------------------------------------------------            
class System(Component):
    """
    Base class for aircraft systems providing core functionality for modeling 
    onboard equipment and subsystems.

    Attributes
    ----------
    tag : str
        Unique identifier for the system component, defaults to 'System'
        
    origin : list
        3D coordinates [x, y, z] defining the system's reference point, 
        defaults to [[0.0, 0.0, 0.0]]
        
    control : Data
        Control system interface parameters, defaults to None
        
    accessories : Data
        Associated auxiliary components and equipment, defaults to None

    Notes
    -----
    The system class serves as the foundation for modeling various aircraft systems:
    
    * Avionics and electronics
    * Environmental control systems
    * Hydraulic systems
    * Fuel systems
    * Auxiliary power units
    
    **Major Assumptions**
    
    * Systems are treated as point masses at their origin
    * Control interfaces are simplified
    * No dynamic response modeling
    
    **Definitions**

    'Origin'
        Reference point for system location and mass properties
        
    'Control Interface'
        Parameters defining how the system interacts with aircraft controls

    See Also
    --------
    RCAIDE.Library.Components.Systems.Avionics
        Implementation for aircraft avionics
    """  
    def __defaults__(self): 
        """
        Sets default values for the system attributes.
        """        
        self.tag         = 'System'
        self.origin      = [[0.0,0.0,0.0]]
        self.control     = None
        self.accessories = None 


# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/induced_drag.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imporst 
from RCAIDE.Framework.Core import Data

# package imports
import numpy as np

# ----------------------------------------------------------------------
#  Induced Drag Aircraft
# ---------------------------------------------------------------------- 
def induced_drag(state,settings,geometry):
    """Determines induced drag for the full aircraft

    Assumptions:
    This function operates in one of three ways:
       -An oswald efficiency is provided. All induced drag, inviscid and viscous, is use calculated using that factor
       -A span efficiency is provided. The inviscid induced drag is calculated from that. Viscous induced drag is
            calculated using the viscous_lift_dependent_drag_factor, K, and the parasite drag.
       -The inviscid induced drag from the the lift calculation for each wing, usually a vortex lattice, is used. Viscous induced drag is
            calculated using the viscous_lift_dependent_drag_factor, K, and the parasite drag.
            
        The last two options do not explicitly provide for the case where the fuselage produces an induced drag directly.

    Source:
    adg.stanford.edu (Stanford AA241 A/B Course Notes)
    http://aerodesign.stanford.edu/aircraftdesign/aircraftdesign.html

    Inputs:
    state.conditions.aerodynamics.
        lift_coefficient                                               [Unitless]
        lift_breakdown.inviscid_wings[wings.*.tag]                     [Unitless]
        drag_breakdown.induced.inviscid                                [Unitless]
        drag_breakdown.induced.inviscid_wings[wings.*.tag]             [Unitless]
        drag_breakdown.parasite[wings.*.tag].parasite_drag_coefficient [Unitless]
        drag_breakdown.parasite[wings.*.tag].reference_area            [m^2]
    settings.oswald_efficiency_factor                                  [Unitless]
    settings.viscous_lift_dependent_drag_factor                        [Unitless]
    settings.span_efficiency                                           [Unitless]
    

    Outputs:
    conditions.aerodynamics.coefficients.drag.induced.
         total                                                         [Unitless]
         viscous                                                       [Unitless]
         oswald_efficiency_factor                                      [Unitless]
         viscous_wings_drag                                            [Unitless]
    
    total_induced_drag                                                 [Unitless]
    

    Properties Used:
    N/A
    """
    # unpack inputs 
    wings         = geometry.wings 
    K             = settings.viscous_lift_dependent_drag_factor
    e_osw         = settings.oswald_efficiency_factor	 
    aero          = state.conditions.aerodynamics.coefficients
    CL            = aero.lift.total
    CDi           = aero.drag.induced.inviscid

    wing_viscous_induced_drags = Data()

    # If the oswald efficiency factor is not specified  
    if e_osw == None:

        # Prime totals
        area                        = 1E-12 
        AR                          = 1E-12 
        total_induced_viscous_drag  = K*aero.drag.parasite.total*(CL**2)

        # Go through each wing, and make calculations
        for wing in wings: 
            AR_wing    = wing.aspect_ratio
            S_wing     = aero.drag.parasite[wing.tag].reference_area  
            if S_wing > area:
                area = S_wing
                AR   = AR_wing
        
        # compute total induced drag 
        total_induced_drag = total_induced_viscous_drag +  CDi  
        
        # Calculate the vehicle level oswald efficiency
        e_osw = (CL**2)/(np.pi*AR*total_induced_drag)

    # If the user specifies a vehicle level oswald efficiency factor
    else: 
        # Find the largest wing, use that for AR
        S  = 1E-12 
        AR = 1E-12 
        for wing in wings:
            if wing.areas.reference>S:
                AR = wing.aspect_ratio
                S  = wing.areas.reference 

        # Calculate the induced drag       
        total_induced_drag = CL **2 / (np.pi*AR*e_osw)
        total_induced_viscous_drag = total_induced_drag - CDi
        
    aero.drag.induced.total                    = total_induced_drag
    aero.drag.induced.viscous                  = total_induced_viscous_drag
    aero.drag.induced.oswald_efficiency_factor = e_osw
    aero.drag.induced.viscous_wings_drag       = wing_viscous_induced_drags 
    
    return 


# RCAIDE/Library/Attributes/Propellants.py
# 
# 
# Created:  Sep 2023, M. Clarke
# Modified: 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from RCAIDE.Framework.Core import Data
from RCAIDE.Library.Components.Mass_Properties import Mass_Properties

# ----------------------------------------------------------------------------------------------------------------------
#  Propellant
# ----------------------------------------------------------------------------------------------------------------------  
class Propellant(Data):
    """Generic class for propellant
    """

    def __defaults__(self):
        """This sets the default values. 
    
        Assumptions:
            None
        
        Source:
            None
        """    
        self.tag                       = 'Propellant'
        self.reactant                  = 'O2'
        self.density                   = 0.0                       # kg/m^3
        self.specific_energy           = 0.0                       # MJ/kg
        self.energy_density            = 0.0                       # MJ/m^3
        self.lower_heating_value       = 0.0                       # MJ/kg
        self.mass_properties           = Mass_Properties()
        self.max_mass_fraction         = Data({'Air' : 0.0, 'O2' : 0.0}) # kg propellant / kg oxidizer
        self.temperatures              = Data()
        self.temperatures.flash        = 0.0                       # K
        self.temperatures.autoignition = 0.0                       # K
        self.temperatures.freeze       = 0.0                       # K
        self.temperatures.boiling      = 0.0                       # K
        
        # Emission Indices 
        self.emission_indices             =  Data() 
        self.emission_indices.Production  = 0
        self.emission_indices.CO2         = 0
        self.emission_indices.H2O         = 0
        self.emission_indices.SO2         = 0
        self.emission_indices.NOx         = 0
        self.emission_indices.Soot        = 0 

        self.global_warming_potential_100           =  Data() 
        self.global_warming_potential_100.CO2       = 0
        self.global_warming_potential_100.H2O       = 0
        self.global_warming_potential_100.SO2       = 0
        self.global_warming_potential_100.NOx       = 0
        self.global_warming_potential_100.Soot      = 0  
        self.global_warming_potential_100.Contrails = 0        


# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/subsonic_parasite_drag_fuselage.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE import
import RCAIDE
from RCAIDE.Framework.Core import Data
from RCAIDE.Library.Methods.Aerodynamics.Common.Drag.compressible_turbulent_flat_plate import compressible_turbulent_flat_plate
from RCAIDE.Library.Methods.Utilities         import Cubic_Spline_Blender   

# python imports 
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#   Parasite Drag Fuselage
# ---------------------------------------------------------------------------------------------------------------------- 
def parasite_drag_fuselage(state,settings,fuselage):
    """Computes the parasite drag due the a fuselage (or boom)

    Assumptions:
        None

    Source:
        Stanford AA241 A/B Course Notes

    Args:
        state.conditions.freestream.
          mach_number                                (numpy.ndarray): mach_number     [Unitless]
          temperature                                (numpy.ndarray): temperature     [K]
          reynolds_number                            (numpy.ndarray): Reynolds number [Unitless]
        settings.fuselage_parasite_drag_form_factor  (numpy.ndarray): form factor     [Unitless]
        geometry.fuselage.       
          areas.front_projected                              (float): projected front area [m^2]
          areas.wetted                                       (float): wetted area          [m^2]
          lengths.total                                      (float): length               [m]
          effective_diameter                                 (float): effective diamater   [m]

    Returns:
        None 
    """
    
    if type(fuselage) == RCAIDE.Library.Components.Fuselages.Blended_Wing_Body_Fuselage: 
        # Store data
        results = Data(
            wetted_area               = 0.0 , 
            reference_area            = 1.0,
            total                     = 0.0,
            skin_friction             = 0.0,
            compressibility_factor    = 0.0,
            reynolds_factor           = 0.0,
            form_factor               = 0.0,
        ) 
        state.conditions.aerodynamics.coefficients.drag.parasite[fuselage.tag] = results
        return  
        
    # unpack inputs   
    Sref          = fuselage.areas.front_projected
    Swet          = fuselage.areas.wetted 
    l_fus         = fuselage.lengths.total
    d_fus         = fuselage.effective_diameter   
    form_factor   = settings.fuselage_parasite_drag_form_factor 
    low_cutoff    = settings.supersonic.fuselage_parasite_drag_begin_blend_mach
    high_cutoff   = settings.supersonic.fuselage_parasite_drag_end_blend_mach  
    Mach          = state.conditions.freestream.mach_number
    T             = state.conditions.freestream.temperature    
    Re            = state.conditions.freestream.reynolds_number 

    # Reynolds number
    Re_fus = Re*(l_fus)
    
    # skin friction coefficient
    cf_fus, k_comp, k_reyn = compressible_turbulent_flat_plate(Re_fus,Mach,T)       
    d_d = float(d_fus)/float(l_fus) 
 
    if np.all((Mach<=1.0) == True): 
        # compute form factor for cylindrical bodies 
        D             = np.zeros_like(Mach)    
        D[Mach < 0.95]  = np.sqrt(1 - (1-Mach[Mach < 0.95]**2) * d_d**2)
        D[Mach >= 0.95] = np.sqrt(1 - d_d**2)
    
        a             = np.zeros_like(Mach)    
        a[Mach < 0.95]  = 2 * (1-Mach[Mach < 0.95]**2) * (d_d**2) *(np.arctanh(D[Mach < 0.95])-D[Mach < 0.95]) / (D[Mach < 0.95]**3)
        a[Mach >= 0.95] = 2  * (d_d**2) *(np.arctanh(D[Mach >= 0.95])-D[Mach >= 0.95]) / (D[Mach >= 0.95]**3)
    
        du_max_u             = np.zeros_like(Mach)    
        du_max_u[Mach < 0.95]  = a[Mach < 0.95] / ( (2-a[Mach < 0.95]) * (1-Mach[Mach < 0.95]**2)**0.5 ) 
        du_max_u[Mach >= 0.95] = a[Mach >= 0.95] / ( (2-a[Mach >= 0.95]) )
        
        k_fus                  = (1 + form_factor*du_max_u)**2 
        fuselage_parasite_drag = k_fus * cf_fus * Swet / Sref
        
    else: 
        
        # supersonic condition  
        D_low        = np.zeros_like(Mach)
        a_low        = np.zeros_like(Mach)
        du_max_u_low = np.zeros_like(Mach)
        
        D_high        = np.zeros_like(Mach)
        a_high        = np.zeros_like(Mach)
        du_max_u_high = np.zeros_like(Mach) 
        k_fus         = np.zeros_like(Mach)
        
        low_inds      = Mach < high_cutoff
        high_inds     = Mach > low_cutoff
        
        D_low[low_inds]        = np.sqrt(1 - (1-Mach[low_inds]**2) * d_d**2)
        a_low[low_inds]        = 2 * (1-Mach[low_inds]**2) * (d_d**2) *(np.arctanh(D_low[low_inds])-D_low[low_inds]) / (D_low[low_inds]**3)
        du_max_u_low[low_inds] = a_low[low_inds] / ( (2-a_low[low_inds]) * (1-Mach[low_inds]**2)**0.5 )
        
        D_high[high_inds]        = np.sqrt(1 - d_d**2)
        a_high[high_inds]        = 2  * (d_d**2) *(np.arctanh(D_high[high_inds])-D_high[high_inds]) / (D_high[high_inds]**3)
        du_max_u_high[high_inds] = a_high[high_inds] / ( (2-a_high[high_inds]) )
        
        spline = Cubic_Spline_Blender(low_cutoff,high_cutoff)
        h00    = lambda M:spline.compute(M)
        
        du_max_u = du_max_u_low*(h00(Mach)) + du_max_u_high*(1-h00(Mach))    
        
        k_fus = (1 + form_factor*du_max_u)**2
    
        fuselage_parasite_drag = k_fus * cf_fus * Swet / Sref
         
    # Store data
    results = Data(
        wetted_area               = Swet   , 
        reference_area            = Sref   , 
        total                     = fuselage_parasite_drag ,
        skin_friction             = cf_fus ,
        compressibility_factor    = k_comp ,
        reynolds_factor           = k_reyn , 
        form_factor               = k_fus  ,
    ) 
    state.conditions.aerodynamics.coefficients.drag.parasite[fuselage.tag] = results        
        
    return  


# RCAIDE/Library/Methods/Propulsors/Turboshaft_Propulsor/append_turboshaft_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_turboshaft_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_turboshaft_conditions(turboshaft,segment):  
    ones_row    = segment.state.ones_row                  
    segment.state.conditions.energy[turboshaft.tag]                               = Conditions()  
    segment.state.conditions.energy[turboshaft.tag].throttle                      = 0. * ones_row(1)     
    segment.state.conditions.energy[turboshaft.tag].commanded_thrust_vector_angle = 0. * ones_row(1)   
    segment.state.conditions.energy[turboshaft.tag].power                         = 0. * ones_row(1)
    segment.state.conditions.energy[turboshaft.tag].fuel_flow_rate                = 0. * ones_row(1)
    segment.state.conditions.energy[turboshaft.tag].inputs                        = Conditions()
    segment.state.conditions.energy[turboshaft.tag].outputs                       = Conditions() 
    segment.state.conditions.noise[turboshaft.tag]                                = Conditions() 
    segment.state.conditions.noise[turboshaft.tag].turboshaft                     = Conditions() 
    segment.state.conditions.noise[turboshaft.tag].turboshaft.core_nozzle         = Conditions()   
    return 


## @ingroup Methods-Geometry-Three_Dimensional
# estimate_naca_4_series_internal_volume.py
#
# Created:  ### ####, M. Vegh
# Modified: Jan 2016, E. Botero

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
import numpy as np

# ----------------------------------------------------------------------
#  Methods
# ----------------------------------------------------------------------

## @ingroup Methods-Geometry-Three_Dimensional
def estimate_wing_internal_volume(wing, m, p): 
    """Computes the volume of a wing with NACA 4-series airfoils.

    Assumptions:
    Wing has constant thickness to chord along the span
    Front spar is at 10%chord and rear spar is at 60% chord

    Source:
    Wikipedia, based on Moran, Jack (2003). An introduction to theoretical and 
      computational aerodynamics. Dover.

    Inputs:
    m                       [-]  percent camber (.1 is 10%)
    p                       [-]  location of max camber (.1 is 10% along chord)
    wing.chords.
      root                  [m]
      tip                   [m]
    wing.taper              [-]
    wing.thickness_to_chord [-]
    wing.spans.projected    [m]
    chord_length            [m]

    Outputs:
    volume                  [m^3] 

    Properties Used:
    N/A
    """          
    #unpack inputs
    t_c   = wing.thickness_to_chord
    taper = wing.taper
    rc    = wing.chords.root
    tc    = wing.chords.tip
    span  = wing.spans.projected

    yc_front_spar_root = m*(.1*rc/p**2.)*(2*p-.1) #front spar is 10%chords
    yc_rear_spar_root  = m*((rc-.6*rc)/(1-p)**2.)*(1+.6-2*p)
    yc_front_spar_tip  = m*(.1*tc/p**2.)*(2*p-.1) 
    yc_rear_spar_tip   = m*((tc-.6*tc)/(1-p)**2.)*(1+.6-2*p)
    
    dy_dx_front_spar = (2.*m/p**2.)*(p-.1) 
    dy_dx_rear_spar  = (2.*m/(1-p)**2.)*(p-.6) 

    theta_front_spar=np.arctan(dy_dx_front_spar)
    theta_rear_spar=np.arctan(dy_dx_rear_spar)
  
    x_front_spar_root  = .1*rc
    x_rear_spar_root   = .6*rc
    yt_front_spar_root = 5*t_c*rc*(.2969*(.1**.5)-.126*.1-.3516*.1**2+.2843*.1**3.-.1015*.1**4.)
    yt_rear_spar_root  = 5*t_c*rc*(.2969*(.6**.5)-.126*.6-.3516*.6**2.+.2843*.6**3.-.1015*.6**4.)
    
    yt_front_spar_tip  = 5*t_c*tc*(.2969*(.1**.5)-.126*.1-.3516*.1**2+.2843*.1**3.-.1015*.1**4.)
    yt_rear_spar_tip   = 5*t_c*tc*(.2969*(.6**.5)-.126*.6-.3516*.6**2.+.2843*.6**3.-.1015*.6**4.)
    
    yu_front_spar_root = yc_front_spar_root+yt_front_spar_root*np.cos(theta_front_spar)
    yl_front_spar_root = yc_front_spar_root-yt_front_spar_root*np.cos(theta_front_spar)
    yu_rear_spar_root  = yc_rear_spar_root+yt_rear_spar_root*np.cos(theta_rear_spar)
    yl_rear_spar_root  = yc_rear_spar_root-yt_rear_spar_root*np.cos(theta_rear_spar)
    
    yu_front_spar_tip  = yc_front_spar_tip+yt_front_spar_tip*np.cos(theta_front_spar)
    yl_front_spar_tip  = yc_front_spar_tip-yt_front_spar_tip*np.cos(theta_front_spar)
    
    yu_rear_spar_tip   = yc_rear_spar_tip+yt_rear_spar_tip*np.cos(theta_rear_spar)
    yl_rear_spar_tip   = yc_rear_spar_tip-yt_rear_spar_tip*np.cos(theta_rear_spar)

    #create line between root and tip points
    slopeu_front_spar  = (yu_front_spar_root-yu_front_spar_tip)/(span/2.)
    slopel_front_spar  = (yl_front_spar_root-yl_front_spar_tip)/(span/2.)
    slopeu_rear_spar   = (yu_rear_spar_root-yu_rear_spar_tip)/(span/2.)
    slopel_rear_spar   = (yl_rear_spar_root-yl_rear_spar_tip)/(span/2.)
    
    yu_front_spar_mid  = (span/4.)*slopeu_front_spar+yu_front_spar_tip
    yl_front_spar_mid  = (span/4.)*slopeu_front_spar+yl_front_spar_tip
    yu_rear_spar_mid   = (span/4.)*slopeu_rear_spar+yu_rear_spar_tip
    yl_rear_spar_mid   = (span/4.)*slopel_rear_spar+yl_rear_spar_tip
    
    slope_chord        = (rc-tc)/(span/2.)
    mid_chord          = slope_chord*(span/4.)+tc
    
    Aroot  = ((yu_front_spar_root-yl_front_spar_root)+(yu_rear_spar_root-yl_rear_spar_root))*(rc*(.6-.1))/2.
    Atip   = ((yu_front_spar_tip-yl_front_spar_tip)+(yu_rear_spar_tip-yl_rear_spar_tip))*(tc*(.6-.1))/2.
    Amid   = ((yu_front_spar_mid-yl_front_spar_mid)+(yu_rear_spar_mid-yl_rear_spar_mid))*(mid_chord*(.6-.1))/2.
    volume = 2*(span/12.)*(Aroot+4*(Amid)+Atip) #integrate a side and multiply by two
    
    return volume


# RCAIDE/Library/Compoments/Wings/Stabilator.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from .Horizontal_Tail    import Horizontal_Tail
from .All_Moving_Surface import All_Moving_Surface
from RCAIDE.Library.Methods.Weights.Moment_of_Inertia.compute_wing_moment_of_inertia import  compute_wing_moment_of_inertia

# ---------------------------------------------------------------------------------------------------------------------- 
#  Stabilator
# ----------------------------------------------------------------------------------------------------------------------     
class Stabilator(Horizontal_Tail, All_Moving_Surface):
    """
    A class representing an all-moving horizontal tail (stabilator) that provides 
    pitch control and longitudinal stability.

    Attributes
    ----------
    tag : str
        Unique identifier for the stabilator, defaults to 'stabilator'
        
    sign_duplicate : float
        Sign convention for duplicate stabilator deflection, defaults to 1.0
        (synchronized deflection for pitch control)

    Notes
    -----
    The stabilator combines the functions of a horizontal stabilizer and elevator 
    by pivoting as a complete surface. It inherits from both Horizontal_Tail and 
    All_Moving_Surface to combine their functionality.

    See Also
    --------
    RCAIDE.Library.Components.Wings.Horizontal_Tail
        Base class providing horizontal tail functionality
    RCAIDE.Library.Components.Wings.All_Moving_Surface
        Base class providing pivoting surface functionality
    RCAIDE.Library.Components.Wings.Control_Surfaces.Elevator
        Alternative pitch control using conventional elevator
    """ 

    def __defaults__(self):
        """
        Sets default values for the stabilator attributes.
        
        Notes
        -----
        See All_Moving_Surface.__defaults__ and Wing.__defaults__ for 
        additional inherited attributes.
        """
        self.tag            = 'stabilator'
        self.sign_duplicate = 1.0

    def moment_of_inertia(wing, center_of_gravity):
        """
        Computes the moment of inertia tensor for the stabilator.

        Parameters
        ----------
        wing : Component
            Wing component data
        center_of_gravity : list
            Reference point coordinates for moment calculation

        Returns
        -------
        ndarray
            3x3 moment of inertia tensor
        """
        I = compute_wing_moment_of_inertia(wing, center_of_gravity) 
        return I  


# RCAIDE/Framework/External_Interfaces/OpenVSP/mach_slices.py
# Created:  May 2021, E. Botero 
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
from . import export_vsp_vehicle

try:
    import vsp as vsp
except ImportError:
    try:
        import openvsp as vsp
    except ImportError:
        # This allows RCAIDE to build without OpenVSP
        pass
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Mach Slices
# ---------------------------------------------------------------------------------------------------------------------- 
def mach_slices(vehicle,mach,angle_of_attack=[0.],number_slices = 99):
    """ This method calculates the volume equivalent area for a vehicle for sonic boom analysis. It will write a VSP
        then slice it up depending on the mach number and angle of attack
    
        Assumptions:
        X_locs is the location where lift values are taken on the x-axis
        AE_x is the lift equivalent area
        
        
        Source:
        N/A
        
        Inputs:
        vehicle             [vehicle]
        mach                [-]
        angle_of_attack     [radians]
        number_slices       [int]


        Outputs:
        X_locs              [m]
        slice_areas         [m^2]
        
        Properties Used:
        N/A
    """       
    

    # Write the vehicle
    export_vsp_vehicle(vehicle,vehicle.tag,write_file=False)
    
    X_locs_all       = []
    slice_areas_all = []
    
    for ii in range(len(mach)):
        
        m   = mach[ii] 
        
        if len(angle_of_attack)>1:
            aoa = angle_of_attack[ii]
        else:
            aoa = angle_of_attack[0]
        
    
        # Calculate the mach angle and adjust for AoA
        mach_angle = np.arcsin(1/m)
        roty = (np.pi/2-mach_angle) + aoa
        
        # Take the components of the X and Z axis to get the slicing plane
        x_component = np.cos(roty)
        z_component = np.sin(roty)
        
        # Now slice it 
        vsp.ComputePlaneSlice( 0, number_slices, vsp.vec3d(x_component[0], 0.0, z_component[0]), True)
        
        # Pull out the areas from the slices
        pslice_results = vsp.FindLatestResultsID("Slice")
        slice_areas    = vsp.GetDoubleResults( pslice_results, "Slice_Area" ) * np.cos(roty)
        vec3d          = vsp.GetVec3dResults(pslice_results, "Slice_Area_Center")
        
        X = []
        Z = []
        
        for v in vec3d:
            X.append(v.x())
            Z.append(v.z())
        
        X = np.array(X)
        Z = np.array(Z)
            
        X_locs = X + Z*np.tan(mach_angle)
        
        if slice_areas[-1]==0.:
            slice_areas = slice_areas[0:-1]
            X_locs      = X_locs[:-1]
            
        # Turn them into arrays
        X_locs      = np.array(X_locs)
        slice_areas = np.array(slice_areas)
        
        # A vectorized Output
        X_locs_all.append(X_locs)
        slice_areas_all.append(slice_areas)

    return X_locs_all, slice_areas_all


# RCAIDE/Methods/Noise/Correlation_Buildup/Engine/secondary_noise_component.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
# Python package imports   
import numpy as np   

# ----------------------------------------------------------------------------------------------------------------------     
#  Secondary Noise Component
# ----------------------------------------------------------------------------------------------------------------------       
def secondary_noise_component(Velocity_primary,theta_s,sound_ambient,Velocity_secondary,Velocity_aircraft,
                              Area_primary,Area_secondary,DSPL_s,EX_s,Str_s):
    """This function calculates the noise contribution of the secondary jet component
    
    Assumptions:
        None

    Source:
        None

    Inputs:
        SPL_s               [dB]
        Velocity_primary    [m/s]
        theta_s             [rad]
        sound_ambient       [dB]
        Velocity_secondary  [m/s]
        Velocity_aircraft   [m/s]
        Area_primary        [m^2]
        Area_secondary      [m^2]
        DSPL_s              [dB]
        EX_s                [-]
        Str_s               [-]

    Outputs: 
        SPL_s               [dB]

    Properties Used:
        N/A 
    
    """

    # Calculation of the velocity exponent
    velocity_exponent = 0.5 * 0.1*theta_s

    # Calculation of the Source Strengh Function (FV)
    FV = ((Velocity_secondary-Velocity_aircraft)/sound_ambient)**velocity_exponent * \
        ((Velocity_secondary+Velocity_aircraft)/sound_ambient)**(1-velocity_exponent)

    # Determination of the noise model coefficients
    Z1 = -18*((1.8*theta_s/np.pi)-0.6)**2
    Z2 = -14-8*((1.8*theta_s/np.pi)-0.6)**3
    Z3 = -0.7
    Z4 = 0.6 - 0.5*((1.8*theta_s/np.pi)-0.6)**2+0.5*(0.6-np.log10(1+Area_secondary/Area_primary))
    Z5 = 51 + 54*theta_s/np.pi - 9*((1.8*theta_s/np.pi)-0.6)**3
    Z6 = 99 + 36*theta_s/np.pi - 15*((1.8*theta_s/np.pi)-0.6)**4 + 5*Velocity_secondary*(Velocity_primary-Velocity_secondary)/(sound_ambient**2) + \
        DSPL_s + EX_s

    # Determination of Sound Pressure Level for the secondary jet component
    SPL_s = (Z1*np.log10(FV)+Z2)*(np.log10(Str_s)-Z3*np.log10(FV)-Z4)**2 + Z5*np.log10(FV) + Z6

    return SPL_s 

